"""Dynamic Network Modes (DyNeMo).

See the `documentation <https://osl-dynamics.readthedocs.io/en/latest/models\
/dynemo.html>`_ for a description of this model.
"""

import os
import logging
from dataclasses import dataclass

import numpy as np
import tensorflow as tf
from tqdm.auto import trange

import osl_dynamics.data.tf as dtf
from osl_dynamics.inference.layers import (
    CovarianceMatricesLayer,
    DiagonalMatricesLayer,
    InferenceRNNLayer,
    KLDivergenceLayer,
    KLLossLayer,
    LogLikelihoodLossLayer,
    MixMatricesLayer,
    MixVectorsLayer,
    ModelRNNLayer,
    NormalizationLayer,
    SampleNormalDistributionLayer,
    SoftmaxLayer,
    VectorsLayer,
)
from osl_dynamics.models import obs_mod
from osl_dynamics.models.inf_mod_base import (
    VariationalInferenceModelBase,
    VariationalInferenceModelConfig,
)
from osl_dynamics.models.mod_base import BaseModelConfig
from osl_dynamics.utils.misc import set_logging_level

_logger = logging.getLogger("osl-dynamics")


@dataclass
class Config(BaseModelConfig, VariationalInferenceModelConfig):
    """Settings for DyNeMo.

    Parameters
    ----------
    model_name : str
        Model name.
    n_modes : int
        Number of modes.
    n_channels : int
        Number of channels.
    sequence_length : int
        Length of sequence passed to the inference network and generative model.

    inference_rnn : str
        RNN to use, either :code:`'gru'` or :code:`'lstm'`.
    inference_n_layers : int
        Number of layers.
    inference_n_units : int
        Number of units.
    inference_normalization : str
        Type of normalization to use. Either :code:`None`, :code:`'batch'` or
        :code:`'layer'`.
    inference_activation : str
        Type of activation to use after normalization and before dropout.
        E.g. :code:`'relu'`, :code:`'elu'`, etc.
    inference_dropout : float
        Dropout rate.
    inference_regularizer : str
        Regularizer.

    model_rnn : str
        RNN to use, either :code:`'gru'` or :code:`'lstm'`.
    model_n_layers : int
        Number of layers.
    model_n_units : int
        Number of units.
    model_normalization : str
        Type of normalization to use. Either None, :code:`'batch'` or
        :code:`'layer'`.
    model_activation : str
        Type of activation to use after normalization and before dropout.
        E.g. :code:`'relu'`, :code:`'elu'`, etc.
    model_dropout : float
        Dropout rate.
    model_regularizer : str
        Regularizer.

    learn_alpha_temperature : bool
        Should we learn :code:`alpha_temperature`?
    initial_alpha_temperature : float
        Initial value for :code:`alpha_temperature`.

    learn_means : bool
        Should we make the mean vectors for each mode trainable?
    learn_covariances : bool
        Should we make the covariance matrix for each mode trainable?
    initial_means : np.ndarray
        Initialisation for mean vectors.
    initial_covariances : np.ndarray
        Initialisation for mode covariances.
        If :code:`diagonal_covariances=True` and full matrices are passed,
        the diagonal is extracted.
    covariances_epsilon : float
        Error added to mode covariances for numerical stability.
    diagonal_covariances : bool
        Should we learn diagonal mode covariances?
    means_regularizer : tf.keras.regularizers.Regularizer
        Regularizer for mean vectors.
    covariances_regularizer : tf.keras.regularizers.Regularizer
        Regularizer for covariance matrices.

    do_kl_annealing : bool
        Should we use KL annealing during training?
    kl_annealing_curve : str
        Type of KL annealing curve. Either :code:`'linear'` or :code:`'tanh'`.
    kl_annealing_sharpness : float
        Parameter to control the shape of the annealing curve if
        :code:`kl_annealing_curve='tanh'`.
    n_kl_annealing_epochs : int
        Number of epochs to perform KL annealing.

    batch_size : int
        Mini-batch size.
    learning_rate : float
        Learning rate.
    lr_decay : float
        Decay for learning rate. Default is 0.1. We use
        :code:`lr = learning_rate * exp(-lr_decay * epoch)`.
    gradient_clip : float
        Value to clip gradients by. This is the :code:`clipnorm` argument
        passed to the Keras optimizer. Cannot be used if :code:`multi_gpu=True`.
    n_epochs : int
        Number of training epochs.
    optimizer : str or tf.keras.optimizers.Optimizer
        Optimizer to use. :code:`'adam'` is recommended.
    loss_calc : str
        How should we collapse the time dimension in the loss?
        Either :code:`'mean'` or :code:`'sum'`.
    multi_gpu : bool
        Should be use multiple GPUs for training?
    strategy : str
        Strategy for distributed learning.
    """

    model_name: str = "DyNeMo"

    # Inference network parameters
    inference_rnn: str = "lstm"
    inference_n_layers: int = 1
    inference_n_units: int = None
    inference_normalization: str = None
    inference_activation: str = None
    inference_dropout: float = 0.0
    inference_regularizer: str = None

    # Model network parameters
    model_rnn: str = "lstm"
    model_n_layers: int = 1
    model_n_units: int = None
    model_normalization: str = None
    model_activation: str = None
    model_dropout: float = 0.0
    model_regularizer: str = None

    # Observation model parameters
    learn_means: bool = None
    learn_covariances: bool = None
    initial_means: np.ndarray = None
    initial_covariances: np.ndarray = None
    diagonal_covariances: bool = False
    covariances_epsilon: float = None
    means_regularizer: tf.keras.regularizers.Regularizer = None
    covariances_regularizer: tf.keras.regularizers.Regularizer = None

    def __post_init__(self):
        self.validate_rnn_parameters()
        self.validate_observation_model_parameters()
        self.validate_alpha_parameters()
        self.validate_kl_annealing_parameters()
        self.validate_dimension_parameters()
        self.validate_training_parameters()

    def validate_rnn_parameters(self):
        if self.inference_n_units is None:
            raise ValueError("Please pass inference_n_units.")

        if self.model_n_units is None:
            raise ValueError("Please pass model_n_units.")

    def validate_observation_model_parameters(self):
        if self.learn_means is None or self.learn_covariances is None:
            raise ValueError("learn_means and learn_covariances must be passed.")

        if self.covariances_epsilon is None:
            if self.learn_covariances:
                self.covariances_epsilon = 1e-6
            else:
                self.covariances_epsilon = 0.0


class Model(VariationalInferenceModelBase):
    """DyNeMo model class.

    Parameters
    ----------
    config : osl_dynamics.models.dynemo.Config
    """

    config_type = Config

    def build_model(self):
        """Builds a keras model."""

        config = self.config

        # ---------- Define layers ---------- #
        data_drop_layer = tf.keras.layers.Dropout(
            config.inference_dropout, name="data_drop"
        )
        inf_rnn_layer = InferenceRNNLayer(
            config.inference_rnn,
            config.inference_normalization,
            config.inference_activation,
            config.inference_n_layers,
            config.inference_n_units,
            config.inference_dropout,
            config.inference_regularizer,
            name="inf_rnn",
        )
        inf_mu_layer = tf.keras.layers.Dense(config.n_modes, name="inf_mu")
        inf_sigma_layer = tf.keras.layers.Dense(
            config.n_modes, activation="softplus", name="inf_sigma"
        )
        theta_layer = SampleNormalDistributionLayer(
            config.theta_std_epsilon,
            name="theta",
        )
        alpha_layer = SoftmaxLayer(
            config.initial_alpha_temperature,
            config.learn_alpha_temperature,
            name="alpha",
        )
        means_layer = VectorsLayer(
            config.n_modes,
            config.n_channels,
            config.learn_means,
            config.initial_means,
            config.means_regularizer,
            name="means",
        )
        if config.diagonal_covariances:
            covs_layer = DiagonalMatricesLayer(
                config.n_modes,
                config.n_channels,
                config.learn_covariances,
                config.initial_covariances,
                config.covariances_epsilon,
                config.covariances_regularizer,
                name="covs",
            )
        else:
            covs_layer = CovarianceMatricesLayer(
                config.n_modes,
                config.n_channels,
                config.learn_covariances,
                config.initial_covariances,
                config.covariances_epsilon,
                config.covariances_regularizer,
                name="covs",
            )
        mix_means_layer = MixVectorsLayer(name="mix_means")
        mix_covs_layer = MixMatricesLayer(name="mix_covs")
        ll_loss_layer = LogLikelihoodLossLayer(config.loss_calc, name="ll_loss")
        theta_drop_layer = tf.keras.layers.Dropout(
            config.model_dropout,
            name="theta_drop",
        )
        mod_rnn_layer = ModelRNNLayer(
            config.model_rnn,
            config.model_normalization,
            config.model_activation,
            config.model_n_layers,
            config.model_n_units,
            config.model_dropout,
            config.model_regularizer,
            name="mod_rnn",
        )
        mod_mu_layer = tf.keras.layers.Dense(config.n_modes, name="mod_mu")
        mod_sigma_layer = tf.keras.layers.Dense(
            config.n_modes, activation="softplus", name="mod_sigma"
        )
        kl_div_layer = KLDivergenceLayer(
            config.theta_std_epsilon, config.loss_calc, name="kl_div"
        )
        kl_loss_layer = KLLossLayer(config.do_kl_annealing, name="kl_loss")

        # ---------- Forward pass ---------- #

        # Encoder
        data = tf.keras.layers.Input(
            shape=(config.sequence_length, config.n_channels), name="data"
        )
        data_drop = data_drop_layer(data)
        inf_rnn = inf_rnn_layer(data_drop)
        inf_mu = inf_mu_layer(inf_rnn)
        inf_sigma = inf_sigma_layer(inf_rnn)
        theta = theta_layer([inf_mu, inf_sigma])
        alpha = alpha_layer(theta)

        # Observation model
        mu = means_layer(data)
        D = covs_layer(data)
        m = mix_means_layer([alpha, mu])
        C = mix_covs_layer([alpha, D])
        ll_loss = ll_loss_layer([data, m, C])

        # Temporal prior
        theta_drop = theta_drop_layer(theta)
        mod_rnn = mod_rnn_layer(theta_drop)
        mod_mu = mod_mu_layer(mod_rnn)
        mod_sigma = mod_sigma_layer(mod_rnn)
        kl_div = kl_div_layer([inf_mu, inf_sigma, mod_mu, mod_sigma])
        kl_loss = kl_loss_layer(kl_div)

        # ---------- Create model ---------- #
        inputs = {"data": data}
        outputs = {"ll_loss": ll_loss, "kl_loss": kl_loss, "theta": theta}
        name = config.model_name
        self.model = tf.keras.Model(inputs=inputs, outputs=outputs, name=name)

    def get_means(self):
        """Get the mode means.

        Returns
        -------
        means : np.ndarary
            Mode means.
        """
        return obs_mod.get_observation_model_parameter(self.model, "means")

    def get_covariances(self):
        """Get the mode covariances.

        Returns
        -------
        covariances : np.ndarary
            Mode covariances.
        """
        return obs_mod.get_observation_model_parameter(self.model, "covs")

    def get_means_covariances(self):
        """Get the mode means and covariances.

        This is a wrapper for :code:`get_means` and :code:`get_covariances`.

        Returns
        -------
        means : np.ndarary
            Mode means.
        covariances : np.ndarray
            Mode covariances.
        """
        return self.get_means(), self.get_covariances()

    def get_observation_model_parameters(self):
        """Wrapper for :code:`get_means_covariances`."""
        return self.get_means_covariances()

    def set_means(self, means, update_initializer=True):
        """Set the mode means.

        Parameters
        ----------
        means : np.ndarray
            Mode means. Shape is (n_modes, n_channels).
        update_initializer : bool
            Do we want to use the passed means when we re-initialize
            the model?
        """
        obs_mod.set_observation_model_parameter(
            self.model,
            means,
            layer_name="means",
            update_initializer=update_initializer,
        )

    def set_covariances(self, covariances, update_initializer=True):
        """Set the mode covariances.

        Parameters
        ----------
        covariances : np.ndarray
            Mode covariances. Shape is (n_modes, n_channels, n_channels).
        update_initializer : bool, optional
            Do we want to use the passed covariances when we re-initialize
            the model?
        """
        obs_mod.set_observation_model_parameter(
            self.model,
            covariances,
            layer_name="covs",
            update_initializer=update_initializer,
            diagonal_covariances=self.config.diagonal_covariances,
        )

    def set_means_covariances(
        self,
        means,
        covariances,
        update_initializer=True,
    ):
        """This is a wrapper for :code:`set_means` and
        :code:`set_covariances`."""
        self.set_means(
            means,
            update_initializer=update_initializer,
        )
        self.set_covariances(
            covariances,
            update_initializer=update_initializer,
        )

    def set_observation_model_parameters(
        self, observation_model_parameters, update_initializer=True
    ):
        """Wrapper for :code:`set_means_covariances`."""
        self.set_means_covariances(
            observation_model_parameters[0],
            observation_model_parameters[1],
            update_initializer=update_initializer,
        )

    def set_regularizers(self, training_dataset):
        """Set the means and covariances regularizer based on the training data.

        A multivariate normal prior is applied to the mean vectors with
        :code:`mu=0`, :code:`sigma=diag((range/2)**2)`. If
        :code:`config.diagonal_covariances=True`, a log normal prior is
        applied to the diagonal of the covariances matrices with :code:`mu=0`,
        :code:`sigma=sqrt(log(2*range))`, otherwise an inverse Wishart prior
        is applied to the covariances matrices with :code:`nu=n_channels-1+0.1`
        and :code:`psi=diag(1/range)`.

        Parameters
        ----------
        training_dataset : tf.data.Dataset or osl_dynamics.data.Data
            Training dataset.
        """
        _logger.info("Setting regularizers")

        training_dataset = self.make_dataset(
            training_dataset, shuffle=False, concatenate=True
        )
        n_sequences, range_ = dtf.get_n_sequences_and_range(training_dataset)
        scale_factor = self.get_static_loss_scaling_factor(n_sequences)

        if self.config.learn_means:
            obs_mod.set_means_regularizer(self.model, range_, scale_factor)

        if self.config.learn_covariances:
            obs_mod.set_covariances_regularizer(
                self.model,
                range_,
                self.config.covariances_epsilon,
                scale_factor,
                self.config.diagonal_covariances,
            )

    def sample_alpha(self, n_samples, theta=None):
        """Uses the model RNN to sample mode mixing factors, :code:`alpha`.

        Parameters
        ----------
        n_samples : int
            Number of samples to take.
        theta : np.ndarray, optional
            Normalized logits to initialise the sampling with.
            Shape must be (sequence_length, n_modes).

        Returns
        -------
        alpha : np.ndarray
            Sampled alpha.
        """
        # Get layers
        model_rnn_layer = self.model.get_layer("mod_rnn")
        mod_mu_layer = self.model.get_layer("mod_mu")
        mod_sigma_layer = self.model.get_layer("mod_sigma")
        alpha_layer = self.model.get_layer("alpha")

        # Normally distributed random numbers used to sample the logits theta
        epsilon = np.random.normal(0, 1, [n_samples + 1, self.config.n_modes]).astype(
            np.float32
        )

        if theta is None:
            # Sequence of the underlying logits theta
            theta = np.zeros(
                [self.config.sequence_length, self.config.n_modes],
                dtype=np.float32,
            )

            # Randomly sample the first time step
            theta[-1] = np.random.normal(size=self.config.n_modes)

        # Sample the mode fixing factors
        alpha = np.empty([n_samples, self.config.n_modes], dtype=np.float32)
        for i in trange(n_samples, desc="Sampling mode time course"):
            # If there are leading zeros we trim theta so that we don't pass
            # the zeros
            trimmed_theta = theta[~np.all(theta == 0, axis=1)][np.newaxis, :, :]

            # Predict the probability distribution function for theta one time
            # step in the future, p(theta|theta_<t) ~ N(mod_mu, sigma_theta_jt)
            model_rnn = model_rnn_layer(trimmed_theta)
            mod_mu = mod_mu_layer(model_rnn)[0, -1]
            mod_sigma = mod_sigma_layer(model_rnn)[0, -1]

            # Shift theta one time step to the left
            theta = np.roll(theta, -1, axis=0)

            # Sample from the probability distribution function
            theta[-1] = mod_mu + mod_sigma * epsilon[i]

            # Calculate the mode mixing factors
            alpha[i] = alpha_layer(mod_mu[np.newaxis, np.newaxis, :])[0, 0]

        return alpha

    def fine_tuning(
        self, training_data, n_epochs=None, learning_rate=None, store_dir="tmp"
    ):
        """Fine tuning the model for each session.

        Here, we train the inference RNN and observation model with the model
        RNN fixed held fixed at the group-level.

        Parameters
        ----------
        training_data : osl_dynamics.data.Data
            Training dataset.
        n_epochs : int, optional
            Number of epochs to train for. Defaults to the value in the
            :code:`config` used to create the model.
        learning_rate : float, optional
            Learning rate. Defaults to the value in the :code:`config` used
            to create the model.
        store_dir : str, optional
            Directory to temporarily store the model in.

        Returns
        -------
        alpha : list of np.ndarray
            Session-specific mixing coefficients.
            Each element has shape (n_samples, n_modes).
        means : np.ndarray
            Session-specific means. Shape is (n_sessions, n_modes, n_channels).
        covariances : np.ndarray
            Session-specific covariances.
            Shape is (n_sessions, n_modes, n_channels, n_channels).
        """
        # Save the group level model
        os.makedirs(store_dir, exist_ok=True)
        self.save_weights(f"{store_dir}/model.weights.h5")

        # Save original training hyperparameters
        original_n_epochs = self.config.n_epochs
        original_learning_rate = self.config.learning_rate
        original_do_kl_annealing = self.config.do_kl_annealing
        self.config.n_epochs = n_epochs or self.config.n_epochs
        self.config.learning_rate = learning_rate or self.config.learning_rate
        self.config.do_kl_annealing = False

        # Layers to fix (i.e. make non-trainable)
        fixed_layers = ["mod_rnn", "mod_mu", "mod_sigma"]

        # Fine tune on sessions
        alpha = []
        means = []
        covariances = []
        with self.set_trainable(fixed_layers, False), set_logging_level(
            _logger, logging.WARNING
        ):
            for i in trange(training_data.n_sessions, desc="Fine tuning"):
                # Train on this session
                with training_data.set_keep(i):
                    self.fit(training_data, verbose=0)
                    a = self.get_alpha(
                        training_data,
                        concatenate=True,
                        verbose=0,
                    )

                # Get inferred parameters
                m, c = self.get_means_covariances()
                alpha.append(a)
                means.append(m)
                covariances.append(c)

                # Reset back to group-level model parameters
                self.load_weights(f"{store_dir}/model.weights.h5")
                self.compile()

        # Reset hyperparameters
        self.config.n_epochs = original_n_epochs
        self.config.learning_rate = original_learning_rate
        self.config.do_kl_annealing = original_do_kl_annealing

        return alpha, np.array(means), np.array(covariances)

    def dual_estimation(
        self,
        training_data,
        n_epochs=None,
        learning_rate=None,
        store_dir="tmp",
    ):
        """Dual estimation to get the session-specific observation model
        parameters.

        Here, we train the observation model parameters (mode means and
        covariances) with the inference RNN and model RNN held fixed at
        the group-level.

        Parameters
        ----------
        training_data : osl_dynamics.data.Data or list of tf.data.Dataset
            Training dataset.
        n_epochs : int, optional
            Number of epochs to train for. Defaults to the value in the
            :code:`config` used to create the model.
        learning_rate : float, optional
            Learning rate. Defaults to the value in the :code:`config` used
            to create the model.
        store_dir : str, optional
            Directory to temporarily store the model in.

        Returns
        -------
        means : np.ndarray
            Session-specific means. Shape is (n_sessions, n_modes, n_channels).
        covariances : np.ndarray
            Session-specific covariances.
            Shape is (n_sessions, n_modes, n_channels, n_channels).
        """
        # Save the group level model
        os.makedirs(store_dir, exist_ok=True)
        self.save_weights(f"{store_dir}/model_weights.weights.h5")

        # Save original training hyperparameters
        original_n_epochs = self.config.n_epochs
        original_learning_rate = self.config.learning_rate
        self.config.n_epochs = n_epochs or self.config.n_epochs
        self.config.learning_rate = learning_rate or self.config.learning_rate

        # Layers to fix (i.e. make non-trainable)
        fixed_layers = [
            "mod_rnn",
            "mod_mu",
            "mod_sigma",
            "inf_rnn",
            "inf_mu",
            "inf_sigma",
            "theta",
            "alpha",
        ]

        # Dual estimation on sessions
        means = []
        covariances = []
        with self.set_trainable(fixed_layers, False):
            if isinstance(training_data, list):
                n_sessions = len(training_data)
            else:
                n_sessions = training_data.n_sessions

            for i in trange(n_sessions, desc="Dual estimation"):
                # Train on this session
                if isinstance(training_data, list):
                    self.fit(training_data[i], verbose=0)
                else:
                    with training_data.set_keep(i):
                        self.fit(training_data, verbose=0)

                # Get inferred parameters
                m, c = self.get_means_covariances()
                means.append(m)
                covariances.append(c)

                # Reset back to group-level model parameters
                self.load_weights(f"{store_dir}/model_weights.weights.h5")
                self.compile()

        # Reset hyperparameters
        self.config.n_epochs = original_n_epochs
        self.config.learning_rate = original_learning_rate

        return np.array(means), np.array(covariances)



"""Base class for models."""

import logging
import os
import re
import warnings
from abc import abstractmethod
from dataclasses import dataclass
from io import StringIO
from contextlib import contextmanager
from packaging import version

import yaml
import numpy as np
import tensorflow as tf
from tensorflow.data import Dataset
from tensorflow.keras import optimizers
from tensorflow.python.distribute.mirrored_strategy import MirroredStrategy
from tqdm.auto import tqdm as tqdm_auto
from tqdm.keras import TqdmCallback

if version.parse(tf.__version__) < version.parse("2.13"):
    from tensorflow.python.distribute.distribution_strategy_context import get_strategy
else:
    from tensorflow.python.distribute.distribute_lib import get_strategy

import osl_dynamics
from osl_dynamics import data
import osl_dynamics.data.tf as dtf
from osl_dynamics.inference import callbacks, initializers
from osl_dynamics.utils.misc import NumpyLoader, get_argument, replace_argument
from osl_dynamics.utils.model import HTMLTable, LatexTable

_logger = logging.getLogger("osl-dynamics")


@dataclass
class BaseModelConfig:
    """Base class for settings for all models."""

    # Model choices
    model_name: str = None
    multiple_dynamics: bool = False

    # Training parameters
    batch_size: int = None
    learning_rate: float = None
    lr_decay: float = 0.1
    gradient_clip: float = None
    n_epochs: int = None
    optimizer: tf.keras.optimizers.Optimizer = "adam"
    loss_calc: str = "mean"
    multi_gpu: bool = False
    strategy: str = None

    # Dimension parameters
    n_modes: int = None
    n_states: int = None
    n_channels: int = None
    sequence_length: int = None

    def validate_training_parameters(self):
        if self.batch_size is None:
            raise ValueError("batch_size must be passed.")
        elif self.batch_size < 1:
            raise ValueError("batch_size must be one or greater.")

        if self.n_epochs is None:
            raise ValueError("n_epochs must be passed.")
        elif self.n_epochs < 1:
            raise ValueError("n_epochs must be one or greater.")

        if self.learning_rate is None:
            raise ValueError("learning_rate must be passed.")
        elif self.learning_rate < 0:
            raise ValueError("learning_rate must be greater than zero.")

        if self.lr_decay < 0:
            raise ValueError("lr_decay must be non-negative.")

        if self.loss_calc not in ["mean", "sum"]:
            raise ValueError("loss_calc must be 'mean' or 'sum'.")

        # Strategy for distributed learning
        if self.multi_gpu:
            self.strategy = MirroredStrategy()
        elif self.strategy is None:
            self.strategy = get_strategy()

    def validate_dimension_parameters(self):
        if self.n_modes is None and self.n_states is None:
            raise ValueError("Either n_modes or n_states must be passed.")

        if self.n_modes is not None:
            if self.n_modes == 1:
                raise ValueError(
                    "n_modes must be two or greater. Consider static analysis."
                )
            if self.n_modes < 1:
                raise ValueError("n_modes must be two or greater.")

        if self.n_states is not None:
            if self.n_states == 1:
                raise ValueError(
                    "n_states must be two or greater. Consider static analysis."
                )
            if self.n_states < 1:
                raise ValueError("n_states must be two or greater.")

        if self.n_channels is None:
            raise ValueError("n_channels must be passed.")
        elif self.n_channels < 1:
            raise ValueError("n_channels must be one or greater.")

        if self.sequence_length is None:
            raise ValueError("sequence_length must be passed.")
        elif self.sequence_length < 1:
            raise ValueError("sequence_length must be one or greater.")


class ModelBase:
    """Base class for all models.

    Acts as a wrapper for a standard Keras model.
    """

    osld_version = None
    config_type = None

    def __init__(self, config):
        self._identifier = np.random.randint(100000)
        self.config = config

        # Build and compile the model
        self.model = None
        with self.config.strategy.scope():
            self.build_model()
            self.compile()

    # Allow access to the keras model attributes
    def __getattr__(self, attr):
        return getattr(self.model, attr)

    @abstractmethod
    def build_model(self):
        """Build a keras model."""
        pass

    def compile(self, optimizer=None):
        """Compile the model.

        Parameters
        ----------
        optimizer : str or tf.keras.optimizers.Optimizer
            Optimizer to use when compiling.
        """

        # Optimizer
        if optimizer is None:
            optimizer = optimizers.get(
                {
                    "class_name": self.config.optimizer.lower(),
                    "config": {
                        "learning_rate": self.config.learning_rate,
                        "clipnorm": self.config.gradient_clip,
                    },
                }
            )

        # Compile
        self.model.compile(optimizer)

        # Add losses to metrics to print during training
        self.add_metrics_for_loss()

    def add_metrics_for_loss(self):
        """Add a metric for each model output loss."""

        # Create metric for each model output that is a loss
        loss_metric = []
        for name in self.output_names:
            if "loss" in name:
                metric = tf.keras.metrics.Mean(name=name)
                loss_metric.append(metric)
        self.model.loss_metric = loss_metric

        # Get the original compute_metrics methods
        old_compute_metrics = self.model.compute_metrics

        # New method for calculating metrics
        def compute_metrics(x, y, y_pred, sample_weight=None):
            metrics = old_compute_metrics(x, y, y_pred, sample_weight)
            for metric in self.loss_metric:
                name = metric.name
                metric.update_state(y_pred[name])
                metrics[name] = metric.result()
            return metrics

        # Override the original Keras method
        self.model.compute_metrics = compute_metrics

    def fit(
        self,
        *args,
        use_tqdm=False,
        tqdm_class=None,
        save_best_after=None,
        checkpoint_freq=None,
        save_filepath=None,
        **kwargs,
    ):
        """Wrapper for the standard keras fit method.

        Adds callbacks and then trains the model.

        Parameters
        ----------
        args : arguments
            Arguments for :code:`keras.Model.fit()`.
        use_tqdm : bool, optional
            Should we use a :code:`tqdm` progress bar instead of the usual
            output from tensorflow.
        tqdm_class : TqdmCallback, optional
            Class for the :code:`tqdm` progress bar.
        save_best_after : int, optional
            Epoch number after which we should save the best model. The best
            model is that which achieves the lowest loss.
        checkpoint_freq : int, optional
            Frequency (in epochs) at which to create checkpoints.
        save_filepath : str, optional
            Path to save the best model to.
        additional_callbacks : list, optional
            List of keras callback objects.
        kwargs : keyword arguments, optional
            Keyword arguments for :code:`keras.Model.fit()`.

        Returns
        -------
        history : history
            The training history.
        """
        # If step_per_epoch is passed, repeat the dataset indefinitely
        steps_per_epoch = get_argument(self.model.fit, "steps_per_epoch", args, kwargs)
        repeat_count = 1 if steps_per_epoch is None else -1

        # If a osl_dynamics.data.Data object has been passed for the x
        # arguments, replace it with a tensorflow dataset
        x = get_argument(self.model.fit, "x", args, kwargs)
        x = self.make_dataset(
            x,
            shuffle=True,
            concatenate=True,
            repeat_count=repeat_count,
        )
        args, kwargs = replace_argument(self.model.fit, "x", x, args, kwargs)

        # Use the number of epochs in the config if it has not been passed
        if get_argument(self.model.fit, "epochs", args, kwargs) is None:
            args, kwargs = replace_argument(
                self.model.fit, "epochs", self.config.n_epochs, args, kwargs
            )

        # If we're saving the model after a particular epoch make sure it's
        # less than the total number of epochs
        if save_best_after is not None:
            epochs = get_argument(self.model.fit, "epochs", args, kwargs)
            if epochs < save_best_after:
                raise ValueError("save_best_after must be less than epochs.")

        # Callbacks to add to the ones the user passed
        additional_callbacks = []

        # Callback to display a progress bar with tqdm
        if use_tqdm:
            tqdm_class = tqdm_class or tqdm_auto
            tqdm_callback = TqdmCallback(verbose=0, tqdm_class=tqdm_class)
            additional_callbacks.append(tqdm_callback)

        # Callback to save the best model after a certain number of epochs
        if save_best_after is not None:
            if save_filepath is None:
                save_filepath = f"/tmp/model_weights/best_{self._identifier}"
            save_best_callback = callbacks.SaveBestCallback(
                save_best_after=save_best_after,
                filepath=save_filepath,
            )
            additional_callbacks.append(save_best_callback)

        if checkpoint_freq is not None:
            if save_filepath is None:
                save_filepath = f"tmp"
            self.save_config(save_filepath)
            checkpoint_callback = callbacks.CheckpointCallback(
                save_freq=checkpoint_freq,
                checkpoint_dir=f"{save_filepath}/checkpoints",
            )
            additional_callbacks.append(checkpoint_callback)

        # Update arguments/keyword arguments to pass to the fit method
        args, kwargs = replace_argument(
            self.model.fit,
            "callbacks",
            additional_callbacks,
            args,
            kwargs,
            append=True,
        )
        if use_tqdm:
            args, kwargs = replace_argument(
                self.model.fit,
                "verbose",
                0,
                args,
                kwargs,
            )

        # Fit model
        history = self.model.fit(*args, **kwargs)

        # Convert history from tensors to float
        history = {
            key: list(map(float, values)) for key, values in history.history.items()
        }

        return history

    def load_weights(self, filepath):
        """Load weights of the model from a file.

        Parameters
        ----------
        filepath : str
            Path to file containing model weights.
        """
        with self.config.strategy.scope():
            with warnings.catch_warnings():
                warnings.filterwarnings(  # suppress optimizer warning
                    "ignore", message="Skipping variable loading for optimizer"
                )
                self.model.load_weights(filepath)

    def reset_weights(self, keep=None):
        """Resets trainable variables in the model to their initial value."""
        initializers.reinitialize_model_weights(self.model, keep=keep)

    def reset(self):
        """Reset the model as if you've built a new model."""
        self.reset_weights()
        self.compile()

    def make_dataset(
        self,
        inputs,
        shuffle=False,
        concatenate=False,
        step_size=None,
        drop_last_batch=False,
        repeat_count=1,
    ):
        """Converts a Data object into a TensorFlow Dataset.

        Parameters
        ----------
        inputs : osl_dynamics.data.Data or str or np.ndarray
            Data object. If a :code:`str` or :np.ndarray: is passed this
            function will first convert it into a Data object.
        shuffle : bool, optional
            Should we shuffle the data?
        concatenate : bool, optional
            Should we return a single TensorFlow Dataset or a list of Datasets.
        step_size : int, optional
            Number of samples to slide the sequence across the dataset.
            Default is no overlap.
        drop_last_batch : bool, optional
            Should we drop the last batch if it is smaller than the batch size?
        repeat_count : int, optional
            Number of times to repeat the dataset.

        Returns
        -------
        dataset : tf.data.Dataset or list
            TensorFlow Dataset (or list of Datasets) that can be used for
            training/evaluating.
        """
        if isinstance(inputs, str) or isinstance(inputs, np.ndarray):
            # str or numpy array -> Data object
            inputs = data.Data(inputs)

        if isinstance(inputs, data.Data):
            # Validation
            if (
                isinstance(self.config.strategy, MirroredStrategy)
                and not inputs.use_tfrecord
            ):
                _logger.warning(
                    "Using a multiple GPUs with a non-TFRecord dataset. "
                    + "This will result in poor performance. "
                    + "Consider using a TFRecord dataset with Data(..., use_tfrecord=True)."
                )

            # Data object -> list of Dataset if concatenate=False
            # or Data object -> Dataset if concatenate=True
            if inputs.use_tfrecord:
                outputs = inputs.tfrecord_dataset(
                    self.config.sequence_length,
                    self.config.batch_size,
                    shuffle=shuffle,
                    concatenate=concatenate,
                    step_size=step_size,
                    drop_last_batch=drop_last_batch,
                    repeat_count=repeat_count,
                    overwrite=True,
                )
            else:
                outputs = inputs.dataset(
                    self.config.sequence_length,
                    self.config.batch_size,
                    shuffle=shuffle,
                    concatenate=concatenate,
                    step_size=step_size,
                    drop_last_batch=drop_last_batch,
                    repeat_count=repeat_count,
                )

        elif isinstance(inputs, Dataset) and not concatenate:
            # Dataset -> list of Dataset if concatenate=False
            outputs = [inputs]

        else:
            outputs = inputs

        return outputs

    def get_training_time_series(
        self,
        training_data,
        prepared=True,
        concatenate=False,
    ):
        """Get the time series used for training from a Data object.

        Parameters
        ----------
        training_data : osl_dynamics.data.Data
            Data object.
        prepared : bool, optional
            Should we return the prepared data? If not, we return the raw data.
        concatenate : bool, optional
            Should we concatenate the data for each session?

        Returns
        -------
        training_data : np.ndarray or list
            Training data time series.
        """
        return training_data.trim_time_series(
            self.config.sequence_length,
            prepared=prepared,
            concatenate=concatenate,
        )

    def get_static_loss_scaling_factor(self, n_sequences):
        """Get scaling factor for static losses.

        When calculating loss, we want to approximate the effect of the
        regularization across the entire training dataset. To do this
        We divide the regularization by the total number of sequences.

        Parameters
        ----------
        n_sequences : int
            Total number of sequences in the training dataset.

        Returns
        -------
        scale_factor : float
            Scale factor for 'static' losses, i.e. those which are not
            time varying.
        """
        scale_factor = 1.0 / n_sequences
        if self.config.loss_calc == "mean":
            scale_factor /= self.config.sequence_length
        return scale_factor

    def summary_string(self):
        """Return a summary of the model as a string.

        This is a modified version of the :code:`keras.Model.summary()` method
        which makes the output easier to parse.
        """
        stringio = StringIO()
        self.model.summary(print_fn=lambda s: stringio.write(s + "\n"))
        return stringio.getvalue()

    def summary_table(self, renderer):
        """Return a summary of the model as a table (HTML or LaTeX).

        Parameters
        ----------
        renderer : str
            Renderer to use. Either :code:`"html"` or :code:`"latex"`.

        Returns
        -------
        table : str
            Summary of the model as a table.
        """

        # Get model.summary() as a string
        summary = self.summary_string()

        renderers = {"html": HTMLTable, "latex": LatexTable}

        # Extract headers
        header_line = summary.splitlines()[2]  # Row with the column headers
        headers = [h.strip() for h in re.split(r"┃", header_line) if h.strip() != ""]

        # Create HTML table
        table = renderers.get(renderer, HTMLTable)(headers)
        for line in summary.splitlines():
            elements = [e.strip() for e in line.split("│")]
            if len(elements) == 1:
                if "params" in elements[0]:
                    elements = re.search(r"(.*? params): (.*?)$", elements[0]).groups()
                    elements = [*elements, "", ""]
                else:
                    continue
            else:
                elements = elements[1:-1]
            table += elements

        return table.output()

    def html_summary(self):
        """Return a summary of the model as an HTML table."""
        return self.summary_table(renderer="html")

    def latex_summary(self):
        """Return a summary of the model as a LaTeX table."""
        return self.summary_table(renderer="latex")

    def _repr_html_(self):
        """Display the model as an HTML table in Jupyter notebooks.

        This is called when you type the variable name of the model in a
        Jupyter notebook. It is unlikely that you will need to call this.
        """
        return self.html_summary()

    def save_config(self, dirname):
        """Saves config object as a .yml file.

        Parameters
        ----------
        dirname : str
            Directory to save :code:`config.yml`.
        """
        os.makedirs(dirname, exist_ok=True)

        config_dict = self.config.__dict__.copy()
        # for serialisability of the dict
        non_serializable_keys = [
            key for key in list(config_dict.keys()) if "regularizer" in key
        ]
        non_serializable_keys.append("strategy")
        for key in non_serializable_keys:
            config_dict[key] = None

        with open(f"{dirname}/config.yml", "w") as file:
            file.write(f"# osl-dynamics version: {osl_dynamics.__version__}\n")
            yaml.dump(config_dict, file)

    def save(self, dirname):
        """Saves config object and weights of the model.

        This is a wrapper for :code:`self.save_config` and
        :code:`self.model.save_weights`.

        Parameters
        ----------
        dirname : str
            Directory to save the :code:`config` object and weights of
            the model.
        """
        self.save_config(dirname)
        self.model.save_weights(f"{dirname}/model.weights.h5")

    @contextmanager
    def set_trainable(self, layers, values):
        """Context manager to temporarily set the :code:`trainable`
        attribute of layers.

        Parameters
        ----------
        layers : str or list of str
            List of layers to set the :code:`trainable` attribute of.
        values : bool or list of bool
            Value to set the :code:`trainable` attribute of the layers to.
        """
        # Validation
        if not isinstance(layers, list):
            layers = [layers]
        if not isinstance(values, list):
            values = [values] * len(layers)
        if len(layers) != len(values):
            raise ValueError(
                "layers and trainable must be the same length, "
                + f"but got {len(layers)} and {len(values)}."
            )

        available_layers = [layer.name for layer in self.layers]
        for i, (layer, value) in enumerate(zip(layers, values)):
            if isinstance(layer, str):
                if layer not in available_layers:
                    raise ValueError(
                        f"Layer {layer} not found in model. Available layers: {available_layers}"
                    )
                layers[i] = self.get_layer(layer)
            elif not isinstance(layer, tf.keras.layers.Layer):
                raise ValueError(
                    f"Layer {layer} is not a string or a Keras layer. "
                    + f"Available layers: {available_layers}"
                )
            if not isinstance(value, bool):
                raise ValueError(f"Value {i} is not a boolean.")

        original_values = [layer.trainable for layer in layers]

        try:
            for layer, trainable in zip(layers, values):
                layer.trainable = trainable
            self.compile()
            yield
        finally:
            for layer, trainable in zip(layers, original_values):
                layer.trainable = trainable
            self.compile()

    @staticmethod
    def load_config(dirname):
        """Load a :code:`config` object from a :code:`.yml` file.

        Parameters
        ----------
        dirname : str
            Directory to load :code:`config.yml` from.

        Returns
        -------
        config : dict
            Dictionary containing values used to create the :code:`config`
            object.
        version : str
            Version used to train the model.
        """

        # Load config dict
        with open(f"{dirname}/config.yml", "r") as file:
            config_dict = yaml.load(file, NumpyLoader)

        # Check what version the model was trained with
        version_comments = []
        with open(f"{dirname}/config.yml", "r") as file:
            for line in file.readlines():
                if "osl-dynamics version" in line:
                    version_comments.append(line)
        if len(version_comments) == 0:
            version = "<1.1.6"
        elif len(version_comments) == 1:
            version = version_comments[0].split(":")[-1].strip()
        else:
            raise ValueError(
                "version could not be read from config.yml. Make sure there "
                + "is only one comment containing the version in config.yml"
            )

        return config_dict, version

    @classmethod
    def load(cls, dirname, from_checkpoint=False, single_gpu=True):
        """Load model from :code:`dirname`.

        Parameters
        ----------
        dirname : str
            Directory where :code:`config.yml` and weights are stored.
        from_checkpoint : bool, optional
            Should we load the model from a checkpoint?
        single_gpu : bool, optional
            Should we compile the model on a single GPU?

        Returns
        -------
        model : Model
            Model object.
        """
        _logger.info(f"Loading model: {dirname}")

        # Load config dict and version from yml file
        config_dict, version = cls.load_config(dirname)

        if single_gpu:
            config_dict["multi_gpu"] = False

        # Create config object
        config = cls.config_type(**config_dict)

        # Create model
        model = cls(config)
        model.osld_version = version

        # Restore model
        if from_checkpoint:
            checkpoint = tf.train.Checkpoint(
                model=model.model, optimizer=model.model.optimizer
            )
            checkpoint.restore(tf.train.latest_checkpoint(f"{dirname}/checkpoints"))
        else:
            cls.load_weights(model, f"{dirname}/model.weights.h5")

        return model

    @property
    def is_multi_gpu(self):
        """Returns True if the model's strategy is MirroredStrategy."""
        return isinstance(self.config.strategy, MirroredStrategy)



"""Multi-Dynamic Network Modes (M-DyNeMo).

See Also
--------
`Example script <https://github.com/OHBA-analysis/osl-dynamics/blob/main\
/examples/simulation/mdynemo_hmm-mvn.py>`_ for training M-DyNeMo on simulated
data (with multiple dynamics).
"""

import logging
from dataclasses import dataclass

import numpy as np
import tensorflow as tf
from tqdm.auto import trange

import osl_dynamics.data.tf as dtf
from osl_dynamics.inference.layers import (
    TFConcatLayer,
    CorrelationMatricesLayer,
    DiagonalMatricesLayer,
    InferenceRNNLayer,
    KLDivergenceLayer,
    KLLossLayer,
    LogLikelihoodLossLayer,
    TFMatMulLayer,
    MixMatricesLayer,
    MixVectorsLayer,
    ModelRNNLayer,
    NormalizationLayer,
    SampleNormalDistributionLayer,
    SoftmaxLayer,
    VectorsLayer,
)
from osl_dynamics.models import obs_mod
from osl_dynamics.models.inf_mod_base import (
    VariationalInferenceModelBase,
    VariationalInferenceModelConfig,
)
from osl_dynamics.models.mod_base import BaseModelConfig

_logger = logging.getLogger("osl-dynamics")


@dataclass
class Config(BaseModelConfig, VariationalInferenceModelConfig):
    """Settings for M-DyNeMo.

    Parameters
    ----------
    model_name : str
        Model name.
    n_modes : int
        Number of modes.
    n_corr_modes : int
        Number of modes for correlation.
        If :code:`None`, then set to :code:`n_modes`.
    n_channels : int
        Number of channels.
    sequence_length : int
        Length of sequence passed to the inference network and generative model.

    inference_rnn : str
        RNN to use, either :code:`'gru'` or :code:`'lstm'`.
    inference_n_layers : int
        Number of layers.
    inference_n_units : int
        Number of units.
    inference_normalization : str
        Type of normalization to use. Either :code:`None`, :code:`'batch'`
        or :code:`'layer'`.
    inference_activation : str
        Type of activation to use after normalization and before dropout.
        E.g. :code:`'relu'`, :code:`'elu'`, etc.
    inference_dropout : float
        Dropout rate.
    inference_regularizer : str
        Regularizer.

    model_rnn : str
        RNN to use, either :code:`'gru'` or :code:`'lstm'`.
    model_n_layers : int
        Number of layers.
    model_n_units : int
        Number of units.
    model_normalization : str
        Type of normalization to use. Either :code:`None`, :code:`'batch'`
        or :code:`'layer'`.
    model_activation : str
        Type of activation to use after normalization and before dropout.
        E.g. :code:`'relu'`, :code:`'elu'`, etc.
    model_dropout : float
        Dropout rate.
    model_regularizer : str
        Regularizer.

    learn_means : bool
        Should we make the mean for each mode trainable?
    learn_stds : bool
        Should we make the standard deviation for each mode trainable?
    learn_corrs : bool
        Should we make the correlation for each mode trainable?
    initial_means : np.ndarray
        Initialisation for the mode means.
    initial_stds : np.ndarray
        Initialisation for mode standard deviations.
    initial_corrs : np.ndarray
        Initialisation for mode correlation matrices.
    stds_epsilon : float
        Error added to mode stds for numerical stability.
    corrs_epsilon : float
        Error added to mode corrs for numerical stability.
    means_regularizer : tf.keras.regularizers.Regularizer
        Regularizer for the mean vectors.
    stds_regularizer : tf.keras.regularizers.Regularizer
        Regularizer for the standard deviation vectors.
    corrs_regularizer : tf.keras.regularizers.Regularizer
        Regularizer for the correlation matrices.

    do_kl_annealing : bool
        Should we use KL annealing during training?
    kl_annealing_curve : str
        Type of KL annealing curve. Either :code:`'linear'` or :code:`'tanh'`.
    kl_annealing_sharpness : float
        Parameter to control the shape of the annealing curve if
        :code:`kl_annealing_curve='tanh'`.
    n_kl_annealing_epochs : int
        Number of epochs to perform KL annealing.

    batch_size : int
        Mini-batch size.
    learning_rate : float
        Learning rate.
    lr_decay : float
        Decay for learning rate. Default is 0.1. We use
        :code:`lr = learning_rate * exp(-lr_decay * epoch)`.
    gradient_clip : float
        Value to clip gradients by. This is the :code:`clipnorm` argument
        passed to the Keras optimizer. Cannot be used if :code:`multi_gpu=True`.
    n_epochs : int
        Number of training epochs.
    optimizer : str or tf.keras.optimizers.Optimizer
        Optimizer to use. :code:`'adam'` is recommended.
    loss_calc : str
        How should we collapse the time dimension in the loss?
        Either :code:`'mean'` or :code:`'sum'`.
    multi_gpu : bool
        Should be use multiple GPUs for training?
    strategy : str
        Strategy for distributed learning.
    """

    model_name: str = "M-DyNeMo"

    # Inference network parameters
    inference_rnn: str = "lstm"
    inference_n_layers: int = 1
    inference_n_units: int = None
    inference_normalization: str = None
    inference_activation: str = None
    inference_dropout: float = 0.0
    inference_regularizer: str = None

    # Model network parameters
    model_rnn: str = "lstm"
    model_n_layers: int = 1
    model_n_units: int = None
    model_normalization: str = None
    model_activation: str = None
    model_dropout: float = 0.0
    model_regularizer: str = None

    # Observation model parameters
    n_corr_modes: int = None
    learn_means: bool = None
    learn_stds: bool = None
    learn_corrs: bool = None
    initial_means: np.ndarray = None
    initial_stds: np.ndarray = None
    initial_corrs: np.ndarray = None
    stds_epsilon: float = None
    corrs_epsilon: float = None
    means_regularizer: tf.keras.regularizers.Regularizer = None
    stds_regularizer: tf.keras.regularizers.Regularizer = None
    corrs_regularizer: tf.keras.regularizers.Regularizer = None
    multiple_dynamics: bool = True

    pca_components: np.ndarray = None

    def __post_init__(self):
        self.validate_rnn_parameters()
        self.validate_observation_model_parameters()
        self.validate_alpha_parameters()
        self.validate_kl_annealing_parameters()
        self.validate_dimension_parameters()
        self.validate_training_parameters()

    def validate_rnn_parameters(self):
        if self.inference_n_units is None:
            raise ValueError("Please pass inference_n_units.")

        if self.model_n_units is None:
            raise ValueError("Please pass model_n_units.")

    def validate_observation_model_parameters(self):
        if (
            self.learn_means is None
            or self.learn_stds is None
            or self.learn_corrs is None
        ):
            raise ValueError("learn_means, learn_stds and learn_corrs must be passed.")

        if self.stds_epsilon is None:
            if self.learn_stds:
                self.stds_epsilon = 1e-6
            else:
                self.stds_epsilon = 0.0

        if self.corrs_epsilon is None:
            if self.learn_corrs:
                self.corrs_epsilon = 1e-6
            else:
                self.corrs_epsilon = 0.0

        if self.pca_components is None:
            self.pca_components = np.eye(self.n_channels)
        self.pca_components = self.pca_components.astype(np.float32)

    def validate_dimension_parameters(self):
        super().validate_dimension_parameters()
        if self.n_corr_modes is None:
            self.n_corr_modes = self.n_modes
            _logger.warning("n_corr_modes is None, set to n_modes.")


class Model(VariationalInferenceModelBase):
    """M-DyNeMo model class.

    Parameters
    ----------
    config : osl_dynamics.models.mdynemo.Config
    """

    config_type = Config

    def build_model(self):
        """Builds a keras model."""

        config = self.config

        # ---------- Define layers ---------- #
        data_drop_layer = tf.keras.layers.Dropout(
            config.inference_dropout, name="data_drop"
        )
        inf_rnn_layer = InferenceRNNLayer(
            config.inference_rnn,
            config.inference_normalization,
            config.inference_activation,
            config.inference_n_layers,
            config.inference_n_units,
            config.inference_dropout,
            config.inference_regularizer,
            name="inf_rnn",
        )
        power_inf_mu_layer = tf.keras.layers.Dense(config.n_modes, name="power_inf_mu")
        power_inf_sigma_layer = tf.keras.layers.Dense(
            config.n_modes, activation="softplus", name="power_inf_sigma"
        )
        power_theta_layer = SampleNormalDistributionLayer(
            config.theta_std_epsilon, name="power_theta"
        )
        fc_inf_mu_layer = tf.keras.layers.Dense(config.n_corr_modes, name="fc_inf_mu")
        fc_inf_sigma_layer = tf.keras.layers.Dense(
            config.n_corr_modes, activation="softplus", name="fc_inf_sigma"
        )
        fc_theta_layer = SampleNormalDistributionLayer(
            config.theta_std_epsilon, name="fc_theta"
        )
        alpha_layer = SoftmaxLayer(
            initial_temperature=1.0,
            learn_temperature=False,
            name="alpha",
        )
        beta_layer = SoftmaxLayer(
            initial_temperature=1.0,
            learn_temperature=False,
            name="beta",
        )
        means_layer = VectorsLayer(
            config.n_modes,
            config.n_channels,
            config.learn_means,
            config.initial_means,
            config.means_regularizer,
            name="means",
        )
        stds_layer = DiagonalMatricesLayer(
            config.n_modes,
            config.n_channels,
            config.learn_stds,
            config.initial_stds,
            config.stds_epsilon,
            config.stds_regularizer,
            name="stds",
        )
        corrs_layer = CorrelationMatricesLayer(
            config.n_corr_modes,
            config.n_channels,
            config.learn_corrs,
            config.initial_corrs,
            config.corrs_epsilon,
            config.corrs_regularizer,
            name="corrs",
        )

        class PCATransformLayer(tf.keras.layers.Layer):
            def __init__(self, pca_components, **kwargs):
                super(PCATransformLayer, self).__init__(**kwargs)
                self.pca_components = tf.Variable(
                    initial_value=tf.convert_to_tensor(
                        pca_components, dtype=tf.float32
                    ),
                    trainable=False,
                )

            def call(self, inputs):
                mu, E, R = inputs
                pca_mu = tf.squeeze(
                    tf.matmul(
                        tf.expand_dims(tf.transpose(self.pca_components), 0),
                        tf.expand_dims(mu, -1),
                    )
                )
                pca_E = tf.matmul(
                    tf.matmul(
                        tf.expand_dims(tf.transpose(self.pca_components), 0),
                        E,
                    ),
                    tf.expand_dims(self.pca_components, 0),
                )
                pca_R = tf.matmul(
                    tf.matmul(
                        tf.expand_dims(tf.transpose(self.pca_components), 0),
                        R,
                    ),
                    tf.expand_dims(self.pca_components, 0),
                )
                return pca_mu, pca_E, pca_R

        pca_layer = PCATransformLayer(config.pca_components, name="pca")
        mix_means_layer = MixVectorsLayer(name="mix_means")
        mix_stds_layer = MixMatricesLayer(name="mix_stds")
        mix_corrs_layer = MixMatricesLayer(name="mix_corrs")
        matmul_layer = TFMatMulLayer(name="cov")
        ll_loss_layer = LogLikelihoodLossLayer(config.loss_calc, name="ll_loss")
        theta_layer = TFConcatLayer(axis=2, name="theta")
        theta_drop_layer = tf.keras.layers.Dropout(
            config.model_dropout,
            name="theta_drop",
        )
        mod_rnn_layer = ModelRNNLayer(
            config.model_rnn,
            config.model_normalization,
            config.model_activation,
            config.model_n_layers,
            config.model_n_units,
            config.model_dropout,
            config.model_regularizer,
            name="mod_rnn",
        )
        power_mod_mu_layer = tf.keras.layers.Dense(config.n_modes, name="power_mod_mu")
        power_mod_sigma_layer = tf.keras.layers.Dense(
            config.n_modes, activation="softplus", name="power_mod_sigma"
        )
        fc_mod_mu_layer = tf.keras.layers.Dense(config.n_corr_modes, name="fc_mod_mu")
        fc_mod_sigma_layer = tf.keras.layers.Dense(
            config.n_corr_modes, activation="softplus", name="fc_mod_sigma"
        )
        fc_kl_div_layer = KLDivergenceLayer(
            config.theta_std_epsilon,
            config.loss_calc,
            name="fc_kl_div",
        )
        kl_div_layer_power = KLDivergenceLayer(
            config.theta_std_epsilon,
            config.loss_calc,
            name="power_kl_div",
        )
        kl_loss_layer = KLLossLayer(config.do_kl_annealing, name="kl_loss")

        # ---------- Forward pass ---------- #

        # Encoder
        data = tf.keras.layers.Input(
            shape=(config.sequence_length, config.n_channels), name="data"
        )
        data_drop = data_drop_layer(data)
        inf_rnn = inf_rnn_layer(data_drop)
        power_inf_mu = power_inf_mu_layer(inf_rnn)
        power_inf_sigma = power_inf_sigma_layer(inf_rnn)
        power_theta = power_theta_layer([power_inf_mu, power_inf_sigma])
        fc_inf_mu = fc_inf_mu_layer(inf_rnn)
        fc_inf_sigma = fc_inf_sigma_layer(inf_rnn)
        fc_theta = fc_theta_layer([fc_inf_mu, fc_inf_sigma])
        alpha = alpha_layer(power_theta)
        beta = beta_layer(fc_theta)

        # Observation model
        mu = means_layer(data)
        E = stds_layer(data)
        R = corrs_layer(data)
        pca_mu, pca_E, pca_R = pca_layer([mu, E, R])
        m = mix_means_layer([alpha, pca_mu])
        G = mix_stds_layer([alpha, pca_E])
        F = mix_corrs_layer([beta, pca_R])
        C = matmul_layer([G, F, G])
        ll_loss = ll_loss_layer([data, m, C])

        # Decoder
        theta = theta_layer([power_theta, fc_theta])
        theta_drop = theta_drop_layer(theta)
        mod_rnn = mod_rnn_layer(theta_drop)
        power_mod_mu = power_mod_mu_layer(mod_rnn)
        power_mod_sigma = power_mod_sigma_layer(mod_rnn)
        fc_mod_mu = fc_mod_mu_layer(mod_rnn)
        fc_mod_sigma = fc_mod_sigma_layer(mod_rnn)
        power_kl_div = kl_div_layer_power(
            [power_inf_mu, power_inf_sigma, power_mod_mu, power_mod_sigma]
        )
        fc_kl_div = fc_kl_div_layer(
            [fc_inf_mu, fc_inf_sigma, fc_mod_mu, fc_mod_sigma],
        )
        kl_loss = kl_loss_layer([power_kl_div, fc_kl_div])

        # ---------- Create model ---------- #
        inputs = {"data": data}
        outputs = {
            "ll_loss": ll_loss,
            "kl_loss": kl_loss,
            "power_theta": power_theta,
            "fc_theta": fc_theta,
        }
        name = config.model_name
        self.model = tf.keras.Model(inputs=inputs, outputs=outputs, name=name)

    def get_means(self):
        """Get the mode means.

        Returns
        -------
        means : np.ndarray
            Mode means. Shape (n_modes, n_channels).
        """
        return obs_mod.get_observation_model_parameter(self.model, "means")

    def get_stds(self):
        """Get the mode standard deviations.

        Returns
        -------
        stds : np.ndarray
            Mode standard deviations. Shape (n_modes, n_channels, n_channels).
        """
        return obs_mod.get_observation_model_parameter(self.model, "stds")

    def get_corrs(self):
        """Get the mode correlations.

        Returns
        -------
        corrs : np.ndarray
            Mode correlations.
            Shape (n_modes, n_channels, n_channels).
        """
        return obs_mod.get_observation_model_parameter(self.model, "corrs")

    def get_means_stds_corrs(self):
        """Get the mode means, standard deviations, correlations.

        This is a wrapper for :code:`get_means`, :code:`get_stds`,
        :code:`get_corrs`.

        Returns
        -------
        means : np.ndarray
            Mode means. Shape is (n_modes, n_channels).
        stds : np.ndarray
            Mode standard deviations.
            Shape is (n_modes, n_channels, n_channels).
        corrs : np.ndarray
            Mode correlations.
            Shape is (n_modes, n_channels, n_channels).
        """
        return self.get_means(), self.get_stds(), self.get_corrs()

    def get_observation_model_parameters(self):
        """Wrapper for :code:`get_means_stds_corrs`."""
        return self.get_means_stds_corrs()

    def set_means(self, means, update_initializer=True):
        """Set the mode means.

        Parameters
        ----------
        means : np.ndarray
            Mode means. Shape is (n_modes, n_channels).
        update_initializer : bool
            Do we want to use the passed parameters when we re-initialize
            the model?
        """
        obs_mod.set_observation_model_parameter(
            self.model,
            means,
            layer_name="means",
            update_initializer=update_initializer,
        )

    def set_stds(self, stds, update_initializer=True):
        """Set the mode standard deviations.

        Parameters
        ----------
        stds : np.ndarray
            Mode standard deviations.
            Shape is (n_modes, n_channels, n_channels) or (n_modes, n_channels).
        update_initializer : bool
            Do we want to use the passed parameters when we re-initialize
            the model?
        """
        obs_mod.set_observation_model_parameter(
            self.model,
            stds,
            layer_name="stds",
            update_initializer=update_initializer,
        )

    def set_corrs(self, corrs, update_initializer=True):
        """Set the mode correlations.

        Parameters
        ----------
        corrs : np.ndarray
            Mode correlations.
            Shape is (n_modes, n_channels, n_channels).
        update_initializer : bool
            Do we want to use the passed parameters when we re-initialize
            the model?
        """
        obs_mod.set_observation_model_parameter(
            self.model,
            corrs,
            layer_name="corrs",
            update_initializer=update_initializer,
        )

    def set_means_stds_corrs(self, means, stds, corrs, update_initializer=True):
        """This is a wrapper for set_means, set_stds, set_corrs."""
        self.set_means(means, update_initializer=update_initializer)
        self.set_stds(stds, update_initializer=update_initializer)
        self.set_corrs(corrs, update_initializer=update_initializer)

    def set_observation_model_parameters(
        self, observation_model_parameters, update_initializer=True
    ):
        """Wrapper for set_means_stds_corrs."""
        self.set_means_stds_corrs(
            observation_model_parameters[0],
            observation_model_parameters[1],
            observation_model_parameters[2],
            update_initializer=update_initializer,
        )

    def set_regularizers(self, training_dataset):
        """Set the regularizers of means, stds and corrs based on the training
        data.

        A multivariate normal prior is applied to the mean vectors with
        :code:`mu=0`, :code:`sigma=diag((range/2)**2)`, a log normal prior is
        applied to the standard deviations with :code:`mu=0`,
        :code:`sigma=sqrt(log(2*range))` and a marginal inverse Wishart prior
        is applied to the functional connectivity matrices with
        :code:`nu=n_channels-1+0.1`.

        Parameters
        ----------
        training_dataset : tf.data.Dataset or osl_dynamics.data.Data
            Training dataset.
        """
        _logger.info("Setting regularizers")

        training_dataset = self.make_dataset(
            training_dataset, shuffle=False, concatenate=True
        )
        n_sequences, range_ = dtf.get_n_sequences_and_range(training_dataset)
        scale_factor = self.get_static_loss_scaling_factor(n_sequences)

        if self.config.learn_means:
            obs_mod.set_means_regularizer(self.model, range_, scale_factor)

        if self.config.learn_stds:
            obs_mod.set_stds_regularizer(
                self.model,
                range_,
                self.config.stds_epsilon,
                scale_factor,
            )

        if self.config.learn_corrs:
            obs_mod.set_corrs_regularizer(
                self.model,
                self.config.n_channels,
                self.config.corrs_epsilon,
                scale_factor,
            )

    def sample_time_courses(self, n_samples):
        """Uses the model RNN to sample mode mixing factors,
        :code:`alpha` and :code:`beta`.

        Parameters
        ----------
        n_samples : int
            Number of samples to take.

        Returns
        -------
        alpha : np.ndarray
            Sampled :code:`alpha`.
        beta : np.ndarray
            Sampled :code:`beta`.
        """
        # Get layers
        model_rnn_layer = self.model.get_layer("mod_rnn")
        power_mod_mu_layer = self.model.get_layer("power_mod_mu")
        power_mod_sigma_layer = self.model.get_layer("power_mod_sigma")
        alpha_layer = self.model.get_layer("alpha")
        fc_mod_mu_layer = self.model.get_layer("fc_mod_mu")
        fc_mod_sigma_layer = self.model.get_layer("fc_mod_sigma")
        beta_layer = self.model.get_layer("beta")
        theta_layer = self.model.get_layer("theta")

        # Normally distributed random numbers used to sample the logits theta
        power_epsilon = np.random.normal(
            0, 1, [n_samples + 1, self.config.n_modes]
        ).astype(np.float32)
        fc_epsilon = np.random.normal(
            0, 1, [n_samples + 1, self.config.n_corr_modes]
        ).astype(np.float32)

        # Initialise sequence of underlying logits theta
        power_theta = np.zeros(
            [self.config.sequence_length, self.config.n_modes],
            dtype=np.float32,
        )
        power_theta[-1] = np.random.normal(size=self.config.n_modes)
        fc_theta = np.zeros(
            [self.config.sequence_length, self.config.n_corr_modes],
            dtype=np.float32,
        )
        fc_theta[-1] = np.random.normal(size=self.config.n_corr_modes)

        # Sample the mode time courses
        alpha = np.empty([n_samples, self.config.n_modes])
        beta = np.empty([n_samples, self.config.n_corr_modes])
        for i in trange(n_samples, desc="Sampling mode time courses"):
            # If there are leading zeros we trim theta so that we don't pass
            # the zeros
            trimmed_power_theta = power_theta[~np.all(power_theta == 0, axis=1)][
                np.newaxis, :, :
            ]
            trimmed_fc_theta = fc_theta[~np.all(fc_theta == 0, axis=1)][
                np.newaxis, :, :
            ]
            trimmed_theta = theta_layer([trimmed_power_theta, trimmed_fc_theta])
            # p(theta|theta_<t) ~ N(mod_mu, sigma_theta_jt)
            model_rnn = model_rnn_layer(trimmed_theta)
            power_mod_mu = power_mod_mu_layer(model_rnn)[0, -1]
            power_mod_sigma = power_mod_sigma_layer(model_rnn)[0, -1]
            fc_mod_mu = fc_mod_mu_layer(model_rnn)[0, -1]
            fc_mod_sigma = fc_mod_sigma_layer(model_rnn)[0, -1]

            # Shift theta one time step to the left
            power_theta = np.roll(power_theta, -1, axis=0)
            fc_theta = np.roll(fc_theta, -1, axis=0)

            # Sample from the probability distribution function
            power_theta[-1] = power_mod_mu + power_mod_sigma * power_epsilon[i]
            fc_theta[-1] = fc_mod_mu + fc_mod_sigma * fc_epsilon[i]

            alpha[i] = alpha_layer(power_theta[-1][np.newaxis, np.newaxis, :])[0, 0]
            beta[i] = beta_layer(fc_theta[-1][np.newaxis, np.newaxis, :])[0, 0]

        return alpha, beta



"""Functions to calculate static properties of time series data."""

import logging

import numpy as np

from osl_dynamics.analysis import spectral

_logger = logging.getLogger("osl-dynamics")


def functional_connectivity(data, conn_type="corr"):
    """Calculate functional connectivity (Pearson correlation).

    Parameters
    ----------
    data : np.ndarray
        Time series data. Shape must be (n_sessions, n_samples, n_channels)
        or (n_samples, n_channels).
    conn_type : str, optional
        What metric should we use? :code:`"corr"` (Pearson correlation) or
        :code:`"cov"` (covariance).

    Returns
    -------
    fc : np.ndarray
        Functional connectivity. Shape is (n_sessions, n_channels, n_channels)
        or (n_channels, n_channels).
    """
    if conn_type not in ["corr", "cov"]:
        raise ValueError(f"conn_type must be 'corr' or 'cov', got: {conn_type}")

    if conn_type == "cov":
        metric = np.cov
    else:
        metric = np.corrcoef

    if isinstance(data, np.ndarray):
        data = [data]
    fc = [metric(d, rowvar=False) for d in data]
    return np.squeeze(fc)


def welch_spectra(
    data,
    sampling_frequency,
    window_length=None,
    step_size=None,
    frequency_range=None,
    return_weights=False,
    standardize=True,
    calc_coh=False,
    keepdims=False,
    n_jobs=1,
):
    """Calculate spectra using Welch's method.

    Wrapper for `spectral.welch_spectra <https://osl-dynamics.readthedocs\
    .io/en/latest/autoapi/osl_dynamics/analysis/spectral/index.html\
    #osl_dynamics.analysis.spectral.welch_spectra>`_ assuming only one
    state is active for all time points.

    Parameters
    ----------
    data : np.ndarray or list
        Time series data. Must have shape (n_sessions, n_samples,
        n_channels) or (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    window_length : int, optional
        Length of the data segment to use to calculate spectra.
        If None, we use :code:`2 * sampling_frequency`.
    step_size : int, optional
        Step size for shifting the window. If None, we use
        :code:`window_length // 2`.
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    standardize : bool, optional
        Should we standardize the data before calculating the spectra?
    calc_coh : bool, optional
        Should we also return the coherence spectra?
    return_weights : bool, optional
        Should we return the weights for session-specific spectra?
        This is useful for calculating a group average.
    keepdims : bool, optional
        Should we enforce a (n_sessions, ...) array is returned
        for :code:`psd` and :code:`coh`? If :code:`False`, we remove any
        dimension of length 1.
    n_jobs : int, optional
        Number of parallel jobs.

    Returns
    -------
    f : np.ndarray
        Frequencies of the power spectra and coherences. Shape is (n_freq,).
    psd : np.ndarray
        Power spectra. Shape is (n_sessions, n_channels, n_freq).
        Any axis of length 1 is removed if :code:`keepdims=False`.
    coh : np.ndarray
        Coherences. Shape is (n_sessions, n_channels, n_channels, n_freq).
        Any axis of length 1 is removed if :code:`keepdims=False`.
        Only returned is :code:`calc_coh=True`.
    w : np.ndarray
        Weighting for session-specific spectra. Only returned if
        :code:`return_weights=True`. Shape is (n_sessions,).
    """
    return spectral.welch_spectra(
        data=data,
        sampling_frequency=sampling_frequency,
        window_length=window_length,
        step_size=step_size,
        frequency_range=frequency_range,
        standardize=standardize,
        calc_coh=calc_coh,
        return_weights=return_weights,
        keepdims=keepdims,
        n_jobs=n_jobs,
    )


def multitaper_spectra(
    data,
    sampling_frequency,
    window_length=None,
    time_half_bandwidth=4,
    n_tapers=7,
    frequency_range=None,
    standardize=True,
    calc_coh=False,
    return_weights=False,
    keepdims=False,
    n_jobs=1,
):
    """Calculate multitaper spectra.

    Wrapper for `spectral.multitaper_spectra <https://osl-dynamics.readthedocs\
    .io/en/latest/autoapi/osl_dynamics/analysis/spectral/index.html\
    #osl_dynamics.analysis.spectral.multitaper_spectra>`_ assuming only one
    state is active for all time points.

    Parameters
    ----------
    data : np.ndarray or list
        Time series data. Must have shape (n_sessions, n_samples,
        n_channels) or (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    window_length : int, optional
        Length of the data segement to use to calculate the multitaper.
    time_half_bandwidth : float, optional
        Parameter to control the resolution of the spectra.
    n_tapers : int, optional
        Number of tapers.
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    standardize : bool, optional
        Should we standardize the data before calculating the multitaper?
    calc_coh : bool, optional
        Should we also return the coherence spectra?
    return_weights : bool, optional
        Should we return the weights for session-specific PSDs?
        Useful for calculating the group average PSD.
    keepdims : bool, optional
        Should we enforce a (n_sessions, n_states, ...) array is returned for
        :code:`psd` and :code:`coh`? If :code:`False`, we remove any dimension
        of length 1.
    n_jobs : int, optional
        Number of parallel jobs.

    Returns
    -------
    f : np.ndarray
        Frequencies of the power spectra and coherences. Shape is (n_freq,).
    psd : np.ndarray
        Power spectra. Shape is (n_sessions, n_channels, n_freq).
        Any axis of length 1 is removed if :code:`keepdims=False`.
    coh : np.ndarray
        Coherences. Shape is (n_sessions, n_channels, n_channels, n_freq).
        Any axis of length 1 is removed if :code:`keepdims=False`.
        Only returned is :code:`calc_coh=True`.
    w : np.ndarray
        Weighting for session-specific spectra. Only returned if
        :code:`return_weights=True`. Shape is (n_sessions,).
    """
    return spectral.multitaper_spectra(
        data=data,
        sampling_frequency=sampling_frequency,
        window_length=window_length,
        time_half_bandwidth=time_half_bandwidth,
        n_tapers=n_tapers,
        frequency_range=frequency_range,
        standardize=standardize,
        calc_coh=calc_coh,
        return_weights=return_weights,
        keepdims=keepdims,
        n_jobs=n_jobs,
    )



"""Functions to manipulate and calculate statistics for inferred mode/state
time courses.

"""

from pathlib import Path

import mne
import numpy as np
import matplotlib.pyplot as plt
from tqdm.auto import trange
from scipy import cluster, spatial, optimize
from sklearn.cluster import AgglomerativeClustering

from osl_dynamics import analysis, array_ops
from osl_dynamics.inference import metrics
from osl_dynamics.utils import plotting
from osl_dynamics.utils.misc import override_dict_defaults


def argmax_time_courses(alpha, concatenate=False, n_modes=None):
    """Hard classifies a time course using an argmax operation.

    Parameters
    ----------
    alpha : list or np.ndarray
        Mode mixing factors or state probabilities. Shape must be
        (n_sessions, n_samples, n_modes) or (n_samples, n_modes).
    concatenate : bool, optional
        If :code:`alpha` is a :code:`list`, should we concatenate the
        time courses?
    n_modes : int, optional
        Number of modes/states there should be. Useful if there are
        modes/states which never activate.

    Returns
    -------
    argmax_tcs : list or np.ndarray
        Argmax time courses. Shape is (n_sessions, n_samples, n_modes)
        or (n_samples, n_modes).
    """
    if isinstance(alpha, list):
        if n_modes is None:
            n_modes = alpha[0].shape[1]
        tcs = [a.argmax(axis=1) for a in alpha]
        tcs = [array_ops.get_one_hot(tc, n_states=n_modes) for tc in tcs]
        if len(tcs) == 1:
            tcs = tcs[0]
        elif concatenate:
            tcs = np.concatenate(tcs)
    elif alpha.ndim == 3:
        if n_modes is None:
            n_modes = alpha.shape[-1]
        tcs = alpha.argmax(axis=2)
        tcs = np.array(
            [array_ops.get_one_hot(tc, n_states=n_modes) for tc in tcs],
        )
        if len(tcs) == 1:
            tcs = tcs[0]
        elif concatenate:
            tcs = np.concatenate(tcs)
    else:
        if n_modes is None:
            n_modes = alpha.shape[1]
        tcs = alpha.argmax(axis=1)
        tcs = array_ops.get_one_hot(tcs, n_states=n_modes)
    return tcs


def gmm_time_courses(
    alpha,
    logit_transform=True,
    standardize=True,
    p_value=None,
    filename=None,
    sklearn_kwargs=None,
    plot_kwargs=None,
):
    """Fit a two-component GMM to time courses to get a binary time course.

    Parameters
    ----------
    alpha : list of np.ndarray or np.ndarray
        Mode time courses. Shape must be (n_sessions, n_samples, n_modes) or
        (n_samples, n_modes).
    logit_transform : bool, optional
        Should we logit transform the mode time course?
    standardize : bool, optional
        Should we standardize the mode time course?
    p_value : float, optional
        Used to determine a threshold. We ensure the data points assigned
        to the 'on' component have a probability of less than :code:`p_value`
        of belonging to the 'off' component.
    filename : str, optional
        Path to directory to plot the GMM fit plots.
    sklearn_kwargs : dict, optional
        Keyword arguments to pass to `sklean.mixture.GaussianMixture \
        <https://scikit-learn.org/stable/modules/generated/\
        sklearn.mixture.GaussianMixture.html>`_.
    plot_kwargs : dict, optional
        Dictionary of keyword arguments to pass to
        `osl_dynamics.utils.plotting.plot_gmm \
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/\
        osl_dynamics/utils/plotting/index.html#osl_dynamics.utils.plotting\
        .plot_gmm>`_.

    Returns
    -------
    gmm_tcs : list of np.ndarray or np.ndarray
        GMM time courses with binary entries. Shape is
        (n_sessions, n_samples, n_modes) or (n_samples, n_modes).
    """
    if plot_kwargs is None:
        plot_kwargs = {}

    if not isinstance(alpha, list):
        alpha = [alpha]

    n_sessions = len(alpha)
    n_modes = alpha[0].shape[1]

    gmm_tcs = []
    gmm_metrics = []
    for sub in trange(n_sessions, desc="Fitting GMMs"):
        # Initialise an array to hold the gmm thresholded time course
        gmm_tc = np.empty(alpha[sub].shape, dtype=int)
        gmm_metric = []

        # Loop over modes
        for mode in range(n_modes):
            a = alpha[sub][:, mode]

            # Fit the GMM
            default_sklearn_kwargs = {"max_iter": 5000, "n_init": 3}
            sklearn_kwargs = override_dict_defaults(
                default_sklearn_kwargs, sklearn_kwargs
            )
            threshold, metrics = analysis.gmm.fit_gaussian_mixture(
                a,
                logit_transform=logit_transform,
                standardize=standardize,
                p_value=p_value,
                sklearn_kwargs=sklearn_kwargs,
                return_statistics=True,
                log_message=False,
            )
            gmm_tc[:, mode] = a > threshold
            gmm_metric.append(metrics)

        # Add to list containing session-specific time courses and
        # component metrics
        gmm_tcs.append(gmm_tc)
        gmm_metrics.append(gmm_metric)

    # Visualise session-specific time courses in one plot per mode
    avg_threshold = [
        np.mean([gmm_metrics[s][m]["threshold"] for s in range(n_sessions)])
        for m in range(n_modes)
    ]
    if filename:
        for mode in range(n_modes):
            # GMM plot filename
            if filename is not None:
                plot_filename = "{fn.parent}/{fn.stem}{mode:0{w}d}{fn.suffix}".format(
                    fn=Path(filename),
                    mode=mode,
                    w=len(str(n_modes)),
                )
            else:
                plot_filename = None

            # session-specific GMM plots per mode
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(7, 4))
            for sub in range(n_sessions):
                metric = gmm_metrics[sub][mode]
                plotting.plot_gmm(
                    metric["data"],
                    metric["amplitudes"],
                    metric["means"],
                    metric["stddevs"],
                    legend_loc=None,
                    ax=ax,
                    **plot_kwargs,
                )
            ax.set_title(f"Averaged Threshold = {avg_threshold[mode]:.3}")
            handles, labels = plt.gca().get_legend_handles_labels()
            label_class = dict(zip(labels, handles))
            ax.legend(label_class.values(), label_class.keys(), loc=1)
            ax.axvline(avg_threshold[mode], color="black", linestyle="--")
            plotting.save(fig, plot_filename)
            plotting.close()

    return gmm_tcs


def correlate_modes(mode_time_course_1, mode_time_course_2):
    """Calculate the correlation matrix between modes in two mode time courses.

    Given two mode time courses, calculate the correlation between each pair of
    modes in the mode time courses. The output for each value in the matrix is
    the value :code:`numpy.corrcoef(mode_time_course_1, \
    mode_time_course_2)[0, 1]`.

    Parameters
    ----------
    mode_time_course_1 : np.ndarray
        Mode time course. Shape must be (n_samples, n_modes).
    mode_time_course_2 : np.ndarray
        Mode time course. Shape must be (n_samples, n_modes).

    Returns
    -------
    correlation_matrix : np.ndarray
        Correlation matrix. Shape is (n_modes, n_modes).
    """
    correlation = np.zeros(
        (mode_time_course_1.shape[1], mode_time_course_2.shape[1]),
    )
    for i, mode1 in enumerate(mode_time_course_1.T):
        for j, mode2 in enumerate(mode_time_course_2.T):
            correlation[i, j] = np.corrcoef(mode1, mode2)[0, 1]
    return correlation


def match_covariances(
    *covariances,
    comparison="rv_coefficient",
    return_order=False,
):
    """Matches covariances.

    Parameters
    ----------
    covariances : tuple of np.ndarray
        Covariance matrices to match.
        Each covariance must be (n_modes, n_channel, n_channels).
    comparison : str, optional
        Either :code:`'rv_coefficient'`, :code:`'correlation'` or
        :code:`'frobenius'`. Default is :code:`'rv_coefficient'`.
    return_order : bool, optional
        Should we return the order instead of the covariances?

    Returns
    -------
    matched_covariances : tuple or list of np.ndarray
        Matched covariances of shape (n_channels, n_channels) or order if
        :code:`return_order=True`.

    Examples
    --------
    Reorder the matrices directly:

    >>> covs1, covs2 = match_covariances(covs1, covs2, comparison="correlation")

    Just get the reordering:

    >>> orders = match_covariances(covs1, covs2, comparison="correlation", return_order=True)
    >>> print(orders[0])  # order for covs1 (always unchanged)
    >>> print(orders[1])  # order for covs2
    """
    # Validation
    for matrix in covariances[1:]:
        if matrix.shape != covariances[0].shape:
            raise ValueError("Matrices must have the same shape.")

    if comparison not in ["frobenius", "correlation", "rv_coefficient"]:
        raise ValueError(
            "Comparison must be 'rv_coefficient', 'correlation' or 'frobenius'."
        )

    # Number of arguments and number of matrices in each argument passed
    n_args = len(covariances)
    n_matrices = covariances[0].shape[0]

    # Calculate the similarity between matrices
    F = np.empty([n_matrices, n_matrices])
    matched_covariances = [covariances[0]]
    orders = [np.arange(covariances[0].shape[0])]
    for i in range(1, n_args):
        for j in range(n_matrices):
            # Find the matrix that is most similar to matrix j
            for k in range(n_matrices):
                if comparison == "frobenius":
                    A = abs(
                        np.diagonal(covariances[i][k]) - np.diagonal(covariances[0][j])
                    )
                    F[j, k] = np.linalg.norm(A)
                elif comparison == "correlation":
                    F[j, k] = -np.corrcoef(
                        covariances[i][k].flatten(), covariances[0][j].flatten()
                    )[0, 1]
                else:
                    F[j, k] = -metrics.pairwise_rv_coefficient(
                        np.array([covariances[i][k], covariances[0][j]])
                    )[0, 1]
        order = optimize.linear_sum_assignment(F)[1]

        # Add the ordered matrix to the list
        matched_covariances.append(covariances[i][order])
        orders.append(order)

    if return_order:
        return orders
    else:
        return tuple(matched_covariances)


def match_vectors(*vectors, comparison="correlation", return_order=False):
    """Matches vectors.

    Parameters
    ----------
    vectors : tuple of np.ndarray
        Sets of vectors to match.
        Each variable must be shape (n_vectors, n_channels).
    comparison : str, optional
        Must be :code:`'correlation'` or :code:`'cosine_similarity'`.
    return_order : bool, optional
        Should we return the order instead of the matched vectors?

    Returns
    -------
    matched_vectors : tuple of np.ndarray
        Set of matched vectors of shape (n_vectors, n_channels)
        or order if :code:`return_order=True`.

    Examples
    --------
    Reorder the vectors directly:

    >>> v1, v2 = match_vectors(v1, v2, comparison="correlation")

    Just get the reordering:

    >>> orders = match_vectors(v1, v2, comparison="correlation", return_order=True)
    >>> print(orders[0])  # order for v1 (always unchanged)
    >>> print(orders[1])  # order for v2
    """
    # Validation
    for vector in vectors[1:]:
        if vector.shape != vectors[0].shape:
            raise ValueError("Vectors must have the same shape.")

    if comparison not in ["correlation", "cosine_similarity"]:
        raise ValueError("Comparison must be 'correlation' or 'cosine_similarity'.")

    # Number of arguments and number of vectors in each argument passed
    n_args = len(vectors)
    n_vectors = vectors[0].shape[0]

    # Calculate the similarity between vectors
    F = np.empty([n_vectors, n_vectors])
    matched_vectors = [vectors[0]]
    orders = [np.arange(vectors[0].shape[0])]
    for i in range(1, n_args):
        for j in range(n_vectors):
            # Find the vector that is most similar to vector j
            for k in range(n_vectors):
                if comparison == "correlation":
                    F[j, k] = -np.corrcoef(vectors[i][k], vectors[0][j])[0, 1]
                elif comparison == "cosine_similarity":
                    F[j, k] = -(
                        1 - spatial.distance.cosine(vectors[i][k], vectors[0][j])
                    )
        order = optimize.linear_sum_assignment(F)[1]

        # Add the ordered vector to the list
        matched_vectors.append(vectors[i][order])
        orders.append(order)

    if return_order:
        return orders
    else:
        return tuple(matched_vectors)


def match_modes(*mode_time_courses, return_order=False):
    """Find correlated modes between mode time courses.

    Given N mode time courses and using the first given mode time course as a
    basis, find the best matches for modes between all of the mode time courses.
    Once found, the mode time courses are returned with the modes reordered so
    that the modes match.

    Given two arrays with columns ABCD and CBAD, both will be returned with
    modes in the order ABCD.

    Parameters
    ----------
    mode_time_courses : list of np.ndarray
        Mode time courses. Each time course must be (n_samples, n_modes).
    return_order : bool, optional
        Should we return the order instead of the mode time courses.

    Returns
    -------
    matched_mode_time_courses : tuple or list of np.ndarray
        Matched mode time courses of shape (n_samples, n_modes) or order
        if :code:`return_order=True`.

    Examples
    --------
    Reorder the modes directly:

    >>> alp1, alp2 = match_modes(alp1, alp2)

    Just get the reordering:

    >>> orders = match_modes(alp1, alp2, return_order=True)
    >>> print(orders[0])  # order for alp1 (always unchanged)
    >>> print(orders[1])  # order for alp2
    """
    # If the mode time courses have different length we only use the
    # first n_samples
    n_samples = min([stc.shape[0] for stc in mode_time_courses])

    # Match time courses based on correlation
    matched_mode_time_courses = [mode_time_courses[0][:n_samples]]
    orders = [np.arange(mode_time_courses[0].shape[1])]
    for mode_time_course in mode_time_courses[1:]:
        correlation = correlate_modes(
            mode_time_courses[0][:n_samples], mode_time_course[:n_samples]
        )
        correlation = np.nan_to_num(
            np.nan_to_num(correlation, nan=np.nanmin(correlation) - 1)
        )
        matches = optimize.linear_sum_assignment(-correlation)
        matched_mode_time_courses.append(mode_time_course[:n_samples, matches[1]])
        orders.append(matches[1])

    if return_order:
        return orders
    else:
        return matched_mode_time_courses


def reduce_state_time_course(state_time_course):
    """Remove states that don't activate from a state time course.

    Parameters
    ----------
    state_time_course: np.ndarray
        State time course. Shape must be (n_samples, n_states).

    Returns
    -------
    reduced_state_time_course: np.ndarray
        Reduced state time course. Shape is (n_samples, n_reduced_states).
    """
    return state_time_course[:, ~np.all(state_time_course == 0, axis=0)]


def fractional_occupancies(state_time_course):
    """Wrapper for `analysis.modes.fractional_occupancies \
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/modes/index.html#osl_dynamics.analysis.modes\
    .fractional_occupancies>`_."""
    return analysis.modes.fractional_occupancies(state_time_course)


def mean_lifetimes(state_time_course, sampling_frequency=None):
    """Wrapper for `analysis.modes.mean_lifetimes \
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/modes/index.html#osl_dynamics.analysis.modes.mean_lifetimes>`_."""
    return analysis.modes.mean_lifetimes(state_time_course, sampling_frequency)


def mean_intervals(state_time_course, sampling_frequency=None):
    """Wrapper for `analysis.modes.mean_intervals \
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/modes/index.html#osl_dynamics.analysis.modes.mean_intervals>`_."""
    return analysis.modes.mean_intervals(state_time_course, sampling_frequency)


def switching_rates(state_time_course, sampling_frequency=None):
    """Wrapper for `analysis.modes.switching_rates \
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/modes/index.html#osl_dynamics.analysis.modes\
    .switching_rates>`_."""
    return analysis.modes.switching_rates(state_time_course, sampling_frequency)


def mean_amplitudes(state_time_course, data):
    """Wrapper for `analysis.modes.mean_amplitudes \
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/modes/index.html#osl_dynamics.analysis.modes\
    .mean_amplitudes>`_."""
    return analysis.modes.mean_amplitudes(state_time_course, data)


def lifetime_statistics(state_time_course, sampling_frequency=None):
    """Wrapper for `analysis.modes.lifetime_statistics \
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/modes/index.html#osl_dynamics.analysis.modes\
    .lifetime_statistics>`_."""
    return analysis.modes.lifetime_statistics(state_time_course, sampling_frequency)


def fano_factor(state_time_course, window_length, sampling_frequency=1.0):
    """Wrapper for `analysis.modes.fano_factor \
    <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
    /analysis/modes/index.html#osl_dynamics.analysis.modes.fano_factor>`_."""
    return analysis.modes.fano_factor(
        state_time_course, window_length, sampling_frequency
    )


def convert_to_mne_raw(
    alpha,
    raw,
    ch_names=None,
    n_embeddings=None,
    n_window=None,
    extra_chans="stim",
    verbose=False,
):
    """Convert a time series to an `MNE Raw \
    <https://mne.tools/stable/generated/mne.io.Raw.html>`_ object.

    Parameters
    ----------
    alpha : np.ndarray
        Time series containing raw data. Shape must be (n_samples, n_modes).
    raw : mne.io.Raw or str
        Raw object to extract info from. If a :code:`str` is passed, it must
        be the path to a fif file containing the Raw object.
    ch_names : list, optional
        Name for each channel. Defaults to :code:`alpha_0, ...,
        alpha_{n_modes-1}`.
    n_embeddings : int, optional
        Number of embeddings that was used to prepare time-delay embedded
        training data.
    n_window : int, optional
        Number of samples used to smooth amplitude envelope data.
    extra_chans : str or list of str, optional
        Extra channel types to add to the Raw object.
    verbose : bool, optional
        Should we print a verbose?

    Returns
    -------
    alpha_raw : mne.io.Raw
        `MNE Raw <https://mne.tools/stable/generated/mne.io.Raw.html>`_ object
        for :code:`alpha`.
    """
    # Validation
    if extra_chans is None:
        extra_chans = []
    if isinstance(extra_chans, str):
        extra_chans = [extra_chans]

    # How many time points from the start of parcellated data should we remove?
    if n_embeddings is not None and n_window is not None:
        raise ValueError("Cannot pass both n_embeddings and n_window.")
    n_trim = n_embeddings or n_window or 1
    n_trim = n_trim // 2

    # Load the Raw object
    if isinstance(raw, str):
        raw = mne.io.read_raw_fif(raw, verbose=verbose)

    # Get time indices excluding bad segments from raw
    _, times = raw.get_data(
        reject_by_annotation="omit", return_times=True, verbose=verbose
    )
    indices = raw.time_as_index(times, use_rounding=True)

    # Remove time points lost due to time delay embedding
    indices = indices[n_trim:]

    # Create an array for the full time series including bad segments
    n_samples = raw.times.shape[0]
    n_channels = alpha.shape[1]
    alpha_ = np.zeros([n_samples, n_channels], dtype=np.float32)

    # Trim the indices we lost when we separate the time series into sequences
    indices = indices[: alpha.shape[0]]

    # Fill in the value for the time series for non-bad segments
    alpha_[indices] = alpha

    # Create info object
    if ch_names is None:
        ch_names = [f"alpha_{ch}" for ch in range(n_channels)]
    alpha_info = mne.create_info(
        ch_names=ch_names,
        ch_types="misc",
        sfreq=raw.info["sfreq"],
        verbose=verbose,
    )

    # Create Raw object
    alpha_raw = mne.io.RawArray(alpha_.T, alpha_info, verbose=verbose)

    # Copy timing info
    alpha_raw.set_meas_date(raw.info["meas_date"])
    alpha_raw.__dict__["_first_samps"] = raw.__dict__["_first_samps"]
    alpha_raw.__dict__["_last_samps"] = raw.__dict__["_last_samps"]
    alpha_raw.__dict__["_cropped_samp"] = raw.__dict__["_cropped_samp"]

    # Copy annotations from raw
    alpha_raw.set_annotations(raw._annotations)

    # Add extra channels
    for extra_chan in extra_chans:
        if extra_chan in raw:
            chan_raw = raw.copy().pick(extra_chan)
            chan_data = chan_raw.get_data()
            chan_info = mne.create_info(
                chan_raw.ch_names,
                raw.info["sfreq"],
                [extra_chan] * chan_data.shape[0],
            )
            chan_raw = mne.io.RawArray(chan_data, chan_info, verbose=verbose)
            alpha_raw.add_channels([chan_raw], force_update_info=True)

    return alpha_raw


def reweight_alphas(alpha, covs):
    """Re-weight mixing coefficients to account for the magnitude of the mode covariances.

    Parameters
    ----------
    alpha : list of np.ndarray or np.ndarray
        Raw mixing coefficients. Shape must be (n_sessions, n_samples, n_modes)
        or (n_samples, n_modes).
    covs : np.ndarray
        Mode covariances. Shape must be (n_modes, n_channels, n_channels).

    Returns
    -------
    reweighted_alpha : list of np.ndarray or np.ndarray
        Re-weighted mixing coefficients. Shape is the same as :code:`alpha`.
    """
    return reweight_mtc(alpha, covs, "covariance")


def reweight_mtc(mtc, params, params_type):
    """Re-weight mixing coefficients to account for the magnitude of
    observation model parameters.

    Parameters
    ----------
    mtc : list of np.ndarray or np.ndarray
        Raw mixing coefficients. Shape must be (n_sessions, n_samples, n_modes)
        or (n_samples, n_modes).
    params : np.ndarray
        Observation model parameters.
        Shape must be (n_modes, n_channels, n_channels).
    params_type : str
        Observation model parameters type. Either 'covariance' or 'correlation'.

    Returns
    -------
    reweighted_mtc : list of np.ndarray
        Re-weighted mixing coefficients. Shape is the same as :code:`mtc`.
    """
    if isinstance(mtc, np.ndarray):
        mtc = [mtc]

    if params_type == "covariance":
        weights = np.trace(params, axis1=1, axis2=2)
    elif params_type == "correlation":
        m, n = np.tril_indices(params.shape[-1], -1)
        weights = np.sum(np.abs(params[:, m, n]), axis=-1)
    else:
        raise ValueError("params_type must be 'covariance' or 'correlation'.")

    reweighted_mtc = [x * weights[np.newaxis, :] for x in mtc]
    reweighted_mtc = [x / np.sum(x, axis=1, keepdims=True) for x in reweighted_mtc]

    if len(reweighted_mtc) == 1:
        reweighted_mtc = reweighted_mtc[0]

    return reweighted_mtc


def average_runs(alpha, n_clusters=None, return_cluster_info=False):
    """Average the state probabilities from different runs using hierarchical clustering.

    Parameters
    ----------
    alpha : list of list of np.ndarray or list of np.ndarray
        State probabilities. Shape must be (n_runs, n_sessions, n_samples,
        n_states) or (n_runs, n_samples, n_states).
    n_clusters : int, optional
        Number of clusters to fit. Defaults to the largest number of states
        in alpha.
    return_cluster_info : bool, optional
        Should we return information describing the clustering?

    Returns
    -------
    average_alpha : list of np.ndarray or np.ndarray
        State probabilities averaged over runs. Shape is (n_sessions, n_states).
    cluster_info : dict
        Clustering info. Only returned if :code:`return_cluster_info=True`.
        This is a dictionary with keys :code:`'correlation'`,
        :code:`'dissimiarity'`, :code:`'ids'` and :code:`'linkage'`.

    See Also
    --------
    S. Alonso and D. Vidaurre, "Towards stability of dynamic FC estimates in
    neuroimaging and electrophysiology: solutions and limits" `bioRxiv (2023): \
    2023-01 <https://www.biorxiv.org/content/10.1101/2023.01.18.524539v2>`_.
    """
    if not isinstance(alpha, list):
        raise ValueError(
            "alpha must be a list of lists (of numpy arrays) or list of numpy arrays."
        )
    if isinstance(alpha[0], np.ndarray):
        alpha = [[a] for a in alpha]

    # Number of runs and length of each session's data
    n_runs = len(alpha)
    n_session_samples = [a.shape[0] for a in alpha[0]]

    # Use the largest number of states as the number of clusters to find
    if n_clusters is None:
        n_clusters = max([a.shape[-1] for a in alpha[0]])

    # Concatenate over arrays, gives (n_runs, n_samples, n_states) array
    alpha = [np.concatenate(a, axis=0) for a in alpha]

    # Turn into a (n_runs * n_states, n_samples) array
    alpha_ = []
    for i in range(n_runs):
        for j in range(alpha[i].shape[-1]):
            alpha_.append(alpha[i][:, j])
    alpha = np.array(alpha_, dtype=np.float32).T

    # Calculate correlation between all pairwise state probability time courses
    corr = np.corrcoef(alpha, rowvar=False)

    # Convert correlation to a dis-similarity measure
    dissimilarity = 1 - corr

    # Hierarchical clustering
    clustering = AgglomerativeClustering(n_clusters, linkage="ward")
    cluster_ids = clustering.fit_predict(dissimilarity)

    # Average alphas in each cluster
    average_alpha = []
    for i in range(n_clusters):
        a = np.mean(alpha[:, cluster_ids == i], axis=-1)
        average_alpha.append(a)
    average_alpha = np.array(average_alpha, dtype=np.float32).T

    # Split average alphas back into session-specific time courses
    average_alpha = np.split(average_alpha, np.cumsum(n_session_samples[:-1]))

    if return_cluster_info:
        # Create a dictionary containing the clustering info
        linkage = cluster.hierarchy.linkage(dissimilarity, method="ward")
        cluster_info = {
            "correlation": corr,
            "dissimilarity": dissimilarity,
            "ids": cluster_ids,
            "linkage": linkage,
        }
        return average_alpha, cluster_info

    else:
        return average_alpha



"""Implemented models."""

import yaml

from osl_dynamics.models import (
    dynemo,
    mdynemo,
    hmm,
    hmm_poi,
    hive,
    dive,
    dyneste,
)
from osl_dynamics.utils import misc

models = {
    "DyNeMo": dynemo.Model,
    "M-DyNeMo": mdynemo.Model,
    "HMM": hmm.Model,
    "HMM-Poisson": hmm_poi.Model,
    "HIVE": hive.Model,
    "DIVE": dive.Model,
    "DyNeStE": dyneste.Model,
}


def load(dirname, single_gpu=True):
    """Load model from dirname.

    Parameters
    ----------
    dirname : str
        Path to directory where the config.yml and weights are stored.
    single_gpu : bool, optional
        Should we compile the model on a single GPU?

    Returns
    -------
    model : DyNeMo model
        Model object.
    """
    with open(f"{dirname}/config.yml", "r") as file:
        config_dict = yaml.load(file, misc.NumpyLoader)

    if "model_name" not in config_dict:
        raise ValueError(
            "Either use a specific `Model.load` method or "
            + "provide a `model_name` field in config"
        )

    try:
        model_type = models[config_dict["model_name"]]
    except KeyError:
        raise NotImplementedError(
            f"{config_dict['model_name']} was not found. "
            + f"Options are {', '.join(models.keys())}"
        )

    return model_type.load(dirname, single_gpu=single_gpu)



"""GLM Permutations base class."""

import logging

import numpy as np
from scipy import stats
from pqdm.processes import pqdm
from tqdm.auto import trange

from osl_dynamics.glm.base import GLM
from osl_dynamics.glm.ols import osl_fit


_logger = logging.getLogger("osl-dynamics")


class Permutation:
    """Base class for permutation tests.

    Parameters
    ----------
    design : osl_dynamics.glm.base.Design
        Design object.
    contrast_indx : int
        Index of the contrast of interest.
    n_perm : int
        Number of permutations.
    perm_type : str, optional
        Type of permutation. Options are 'sign_flip' and 'row_shuffle'.
        If None, it will be determined based on the feature types and contrast type.
    n_jobs : int, optional
        Number of jobs to run in parallel.
    """

    def __init__(self, design, contrast_indx, n_perm, perm_type=None, n_jobs=1):
        self.glm = GLM(design)
        self.contrast_indx = contrast_indx
        self.c = self.glm.c[self.contrast_indx][None, :]
        self.n_perm = n_perm
        self.n_jobs = n_jobs
        self.perm_type = self._validate_perm_type(perm_type)

    def permute_X(self):
        """Permute the design matrix based on the perm_type.

        Returns
        -------
        X_copy : np.ndarray
            Permuted design matrix. Shape is :code:`(n_samples, n_features)`.
        """
        X_copy = self.glm.X.copy()
        permute_indx = self._get_permute_feature_indx()
        if self.perm_type == "sign_flip":
            # Randomly flip the sign of the features
            signs = np.random.choice([-1, 1], self.glm.n_samples)
            X_copy[:, permute_indx] *= signs[:, None]
        else:
            # Randomly shuffle the rows
            row_indx = np.random.permutation(self.glm.n_samples)
            X_copy[:, permute_indx] = X_copy[np.ix_(row_indx, permute_indx)]

        return X_copy

    def fit(self, y):
        """Fit the GLM with unpermuted data and run permutations.

        Parameters
        ----------
        y : np.ndarray
            Target variable. Shape is :code:`(n_samples, *target_dims)`.
        """
        self.glm.fit(y)
        y_flatten = np.reshape(y, (self.glm.n_samples, -1))
        # Build keyword arguments for parallel processing
        kwargs = []
        for _ in range(self.n_perm):
            kwargs.append(
                {
                    "X": self.permute_X(),
                    "y": y_flatten,
                    "contrasts": self.c,
                }
            )

        # Run permutations
        if len(kwargs) == 1:
            _logger.info(
                "Running permutations on contrast "
                f"{self.glm.contrast_names[self.contrast_indx]}."
            )
            results = [osl_fit(**kwargs[0])]
        elif self.n_jobs == 1:
            _logger.info(
                f"Running on contrast {self.glm.contrast_names[self.contrast_indx]} "
                f"permutations with {self.n_jobs} jobs."
            )
            results = []
            for i in trange(self.n_perm, desc="Running permutations"):
                results.append(osl_fit(**kwargs[i]))
        else:
            _logger.info(
                f"Running on contrast {self.glm.contrast_names[self.contrast_indx]} "
                f"permutations with {self.n_jobs} jobs."
            )
            results = pqdm(
                kwargs,
                osl_fit,
                argument_type="kwargs",
                n_jobs=self.n_jobs,
                desc="Running permutations",
            )

        # Unpack results
        null_copes, null_tstats = [], []
        for result in results:
            _, copes, varcopes = result
            null_copes.append(copes)
            null_tstats.append(self.glm.get_tstats(copes, varcopes))

        self.null_copes = np.reshape(null_copes, (self.n_perm, *self.glm.target_dims))
        self.null_tstats = np.reshape(null_tstats, (self.n_perm, *self.glm.target_dims))

    def _get_permute_feature_indx(self):
        """Get the indices of the features to permute."""
        return np.where(self.glm.c[self.contrast_indx] != 0.0)[0]

    def _validate_perm_type(self, perm_type):
        if perm_type is not None:
            if perm_type not in ["sign_flip", "row_shuffle"]:
                raise ValueError(
                    f"perm_type must be 'sign_flip' or 'row_shuffle', got {perm_type}"
                )
            return perm_type

        permute_indx = self._get_permute_feature_indx()
        feature_types = np.array(self.glm.feature_types)[permute_indx]
        feature_type = np.unique(feature_types)
        contrast_type = self.glm.contrast_types[self.contrast_indx]

        if len(feature_type) > 1:
            raise ValueError(
                "Cannot determine perm_type when feature types are mixed. "
                f"Got {feature_type}"
            )

        if feature_type == "constant":
            return "sign_flip"

        if feature_type == "categorical":
            if contrast_type == "differential":
                return "row_shuffle"
            return "sign_flip"

        return "row_shuffle"

    @property
    def copes(self):
        """Contrast Of Parameter Estimates."""
        return self.glm.copes[self.contrast_indx]

    @property
    def tstats(self):
        """T-stats of the contrast of interest."""
        return self.glm.tstats[self.contrast_indx]

    def summary(self):
        """Print summary of the permutation test."""
        sum = self.glm.summary()
        sum["n_perm"] = self.n_perm
        sum["perm_type"] = self.perm_type
        sum["contrast_indx"] = self.contrast_indx
        sum["n_jobs"] = self.n_jobs
        return sum


class MaxStatPermutation(Permutation):
    """
    Max statistic permutation test.

    Parameters
    ----------
    design : osl_dynamics.glm.base.Design
        Design object.
    contrast_indx : int
        Index of the contrast of interest.
    n_perm : int
        Number of permutations.
    perm_type : str, optional
        Type of permutation. Options are 'sign_flip' and 'row_shuffle'.
        If None, it will be determined based on the feature types and
        contrast type.
    n_jobs : int, optional
        Number of jobs to run in parallel.
    """

    def fit(self, y):
        """Fit the GLM with unpermuted data and run permutations.

        Parameters
        ----------
        y : np.ndarray
            Target variable. Shape is :code:`(n_samples, *target_dims)`.
        """
        super().fit(y)
        pool_axis = tuple(range(1, len(self.glm.target_dims) + 1))
        self.null_max_copes = np.nanmax(np.abs(self.null_copes), axis=pool_axis)
        self.null_max_tstats = np.nanmax(np.abs(self.null_tstats), axis=pool_axis)

    def get_pvalues(self, metric="copes"):
        """Get p-values.

        Parameters
        ----------
        metric : str, optional
            Metric to compute p-values. Options are 'copes' and 'tstats'.

        Returns
        -------
        pvalues : np.ndarray
            P-values. Shape is :code:`(*target_dims)`.
        """
        if metric == "copes":
            obs_stat = np.abs(self.glm.copes[self.contrast_indx])
            percentiles = stats.percentileofscore(self.null_max_copes, obs_stat)
        elif metric == "tstats":
            obs_stat = np.abs(self.glm.tstats[self.contrast_indx])
            percentiles = stats.percentileofscore(self.null_max_tstats, obs_stat)
        else:
            raise ValueError(f"metric must be 'copes' or 'tstats', got {metric}")

        return 1 - percentiles / 100
